from enum import Enum
import inspect
import re
import sys
from functools import wraps
from typing import Any, Callable

import pandas as pd

from prism._utils.exceptions import PrismTypeError, PrismValueError, PrismAuthError
from .._common import const


def handle_jupyter_exception(func):
    @wraps(func)
    def showtraceback(*args, **kwargs):
        # extract exception type, value and traceback
        etype, evalue, tb = sys.exc_info()
        customType = "Prism" in etype.__name__
        if customType:
            print(evalue)
            return
        else:
            # otherwise run the original hook
            value = func(*args, **kwargs)
            return value
    return showtraceback


def handle_sys_exception(exc_type, exc_value, exc_traceback):
    customType = "Prism" in exc_type.__name__
    if customType:
        print(exc_value)
        return
    else:
        # otherwise run the original hook
        return sys.__excepthook__(exc_type, exc_value, exc_traceback)


def _validate_args(function):
    signature = inspect.signature(function)

    @wraps(function)
    def type_checker(*args, **kwargs):
        try:
            bound_args = signature.bind(*args, **kwargs)
            bound_args.apply_defaults()
            bound_args_dict = dict(bound_args.arguments)


            for param_name, param_info in signature.parameters.items():
                if param_name == "self":
                    continue

                arg = bound_args_dict[param_name]
                if arg is None:
                    continue

                param_type = param_info.annotation

                check_argument_type(arg, param_type, param_name)
                arg = apply_validation_rules(
                    arg, param_name, function.__name__)

                bound_args_dict[param_name] = arg

            function_specific_validation(bound_args_dict, function.__name__)

            return function(**bound_args_dict)
        except TypeError as e:
            raise PrismTypeError(f"{e}")
    return type_checker


def function_specific_validation(args_dict, function_name):
    if function_name == "fillna":
        v, m = args_dict.get("value"), args_dict.get("method")
        if not (v is not None) ^ (m is not None):
            raise PrismValueError("Must specify one fill 'value' or 'method'")
    elif function_name == "clip":
        l, u = args_dict.get("lowerbound"), args_dict.get("upperbound")
        if (l is None) and (u is None):
            raise PrismValueError("Must specify at leat one of 'lowerbound' or 'upperbound'")
    elif function_name == "log_n" and args_dict["n"] <= 0:
        raise Exception("Base must be positive")


def check_argument_type(arg, param_type, param_name):
    if param_type is inspect.Parameter.empty:
        return

    if isinstance(param_type, type):
        if issubclass(param_type, Enum):
            try:
                arg = param_type(arg)
            except Exception as e:
                raise PrismValueError(f"{e}", valid_list=param_type) from e
        elif not isinstance(arg, param_type):
            raise PrismTypeError(
                f"Type of {param_name} is {type(arg)} and not {param_type}")
    elif not any(isinstance(arg, t) for t in param_type.__args__):
        raise PrismTypeError(
            f"Type of {param_name} is {type(arg)} and not in the Union of {param_type.__args__}")


def apply_validation_rules(arg, param_name: str, function_name: str):
    # if param_name equals function name
    if param_name in validation_functions:
        validation_function = validation_functions[param_name]
        return validation_function(arg, function_name)

    # if param_name matches custom pattern
    for _, (pattern_checker, validation_function) in custom_validation_functions.items():
        if pattern_checker(param_name):
            return validation_function(arg)

    # when there is no validation rule to check
    return arg


def startdate_param_validator(arg, _):
    try:
        date = pd.to_datetime(arg)
    except Exception as e:
        raise PrismValueError(
            f'Unknown string format. Cannot parse "{arg}" to a date') from e

    assert (date >= const.BEGINNING_DATE) & (
        date <= const.ACTIVE_DATE), "Not a valid date."
    return arg


def enddate_param_validator(arg, _):
    return startdate_param_validator(arg, _)


def setting_param_validator(arg, _):
    if const.PreferenceType is None:
        raise PrismAuthError("Please Login First")
    if (arg not in const.PreferenceType) and (arg != ""):
        raise PrismValueError(
            "Invalid preference keyword is given. ", valid_list=const.PreferenceType)
    return arg


def settings_param_validator(arg, _):
    if const.PreferenceType is None:
        raise PrismAuthError("Please Login First")
    for k in arg.keys():
        if k not in const.PreferenceType:
            raise PrismValueError(
                "Invalid preference key is given. ", valid_list=const.PreferenceType)
    return arg


def frequency_validator(frequency: str, function_name):
    if function_name in ["resample", "get_universe"]:
        business_pattern = "b?"
    else:
        business_pattern = ""
    days_str = "|".join(const.DaysRepr)
    day_pattern = fr"([1-9][0-9]{{0,3}})?{business_pattern}d"
    week_pattern = fr"([1-9][0-9]{{0,2}})?{business_pattern}w(-(0[1-7]|[1-7]|{days_str}))?"
    month_pattern = fr"([1-9][0-9]{{0,1}})?{business_pattern}(m|q|sa)(-(0[1-9]|[12]\d|3[01]|[1-9]))?"
    yearly_pattern = fr"[1-9]?{business_pattern}a(-(0?[1-9]|[1][0-2])\/(0[1-9]|[12]\d|3[01]|[1-9]))?"

    frequency_pattern = fr"({day_pattern})|({week_pattern})|({month_pattern})|{yearly_pattern}"

    if re.fullmatch(frequency_pattern, frequency.lower()) is None:
        raise PrismValueError("Not a valid frequency format - Please refer to the User Guide")
    return frequency


def sample_period_validator(period: int, _):
    # period_pattern = fr"([1-9][0-9]{{0,3}})?(D|W|M|Q|SA|A)"

    # if re.fullmatch(period_pattern, period) is None:
    #     raise PrismValueError("Not a valid sample period format - Please refer to the User Guide")
    if not isinstance(period, int):
        raise PrismValueError("Not a valid sample period format - Please refer to the User Guide")
    return period

# if param_name equals dict's key name
# TODO: refactor this code using eval
validation_functions: dict[str, Callable[[Any, str], Any]] = {
    "startdate": startdate_param_validator,
    "enddate": enddate_param_validator,
    "setting": setting_param_validator,
    "settings": settings_param_validator,
    "sample_period": sample_period_validator,
    "frequency": frequency_validator,
}


# custom validators
def custom_universename_validator(arg):
    regex = re.compile("[@_!#$%^&*()<>?/|}{~:]`\"'")
    if isinstance(arg, list):
        for a in arg:
            if regex.search(a) is not None:
                raise PrismValueError(
                    "universename not to include special characters")
    else:
        if regex.search(arg) is not None:
            raise PrismValueError(
                "universename not to include special characters")
    return arg


def custom_min_periods_validator(arg):
    if arg < 1:
        raise PrismValueError("min_periods cannot be less than 1")
    return arg


# uses custum pattern to check param_name
custom_validation_functions: dict[str, tuple[Callable[[str], bool], Callable[[Any], Any]]] = {
    "universename": (lambda param_name: "universename" in param_name, custom_universename_validator),
    "min_periods": (lambda param_name: "min_periods" in param_name, custom_min_periods_validator)
}


def get_sm_attributevalue(attribute: str):
    if const.SMValues is None:
        raise PrismAuthError(f"Please Login First")
    smattributes_lower = {
        a.lower().replace(' ', '').replace('_', ''): const.SMValues[a]
        for a in const.SMValues.keys()
    }
    smattributes_lower.update({
        a.lower().replace(' ', '').replace('_', ''): a
        for a in const.SMValues.values()
    })
    attribute_lower = attribute.lower().replace(' ', '').replace('_', '')
    smattributevalue = smattributes_lower.get(attribute_lower)
    if smattributevalue is None:
        raise PrismValueError(
            f"{attribute} is not a valid Security Master attribute",
            valid_list=list(const.SMValues.keys())
        )
    return smattributevalue


from typing import Union

import pytest

from prism._utils.validate_utils import _validate_args
from prism._utils.exceptions import PrismTypeError, PrismValueError


@_validate_args
def filter_universe(condition: list, universename: Union[str, list]):
    return True


condition = [{"attribute": 'CIQ primary',
              'value': ['primary'], 'operator': 'AND'}]


def test_universename_param():
    # valid universename
    assert filter_universe(condition, 'APAC_primary')
    assert filter_universe(condition, ['APAC_primary'])
    assert filter_universe(condition, [
                           'APAC_primary"', 'Hello! World!\n', 'THERE # is ? special @ char !! inside.'])
    assert filter_universe(condition, "This is OK`\"'To be inside")
    assert filter_universe(condition, "This is OK#\"'To be inside")
    assert filter_universe(condition, "This is OK#`'To be inside")


def test_wrong_universename_param():
    # wrong value (regex test)
    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "@`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "_`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "!`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "#`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "$`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "%`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "^`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "&`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "*`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "(`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, ")`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "<`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, ">`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "?`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "|`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "{`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "}`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "~`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, ":`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "APAC_!`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, "HELLO#`\"'")

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, ['APAC_primary', "SPECIAL#`\"'"])

    with pytest.raises(PrismValueError):
        assert filter_universe(condition, ["INSIDE@@@`\"'AA", 'APAC_primary'])

    # wrong type
    with pytest.raises(PrismTypeError):
        assert filter_universe(condition=condition, universename=1)

    with pytest.raises(PrismTypeError):
        assert filter_universe(condition=1)

    with pytest.raises((TypeError, PrismTypeError)):
        assert filter_universe(condition=condition)


def test_function_param_contains_univername():
    @_validate_args
    def sample_function(param_universename: str | list):
        return True

    assert sample_function("my universe")

    with pytest.raises(PrismValueError):
        assert sample_function("APAC#`\"'_primary")

    with pytest.raises(PrismValueError):
        assert sample_function(["@@@`\"'with special"])



import datetime
from datetime import timedelta

from sqlalchemy.sql.schema import MetaData
from sqlalchemy import (
    Table,
    Column,
    Integer,
    String,
    ForeignKey,
    SmallInteger,
    TypeDecorator,
    DateTime as DT,
    Interval as IT,
    Boolean,
    Index
)

class PrismTimeDelta(TypeDecorator):
    impl = IT
    cache_ok = True

    def process_literal_param(self, value, _) -> None:
        v = str(value).replace(", 0:00:00", "")
        return "interval '" + v + "'"


class PrismDateTime(TypeDecorator):
    impl = DT
    cache_ok = True

    def coerce_compared_value(self, op, value):
        if isinstance(value, timedelta):
            return PrismTimeDelta()
        else:
            return self

    def process_literal_param(self, value, _) -> None:
        return str(value).replace(" 00:00:00", "")


metadata = MetaData()

dataqueries = Table(
    'dataqueries',
    metadata,
    Column('dataqueryid', Integer, primary_key=True, autoincrement=True),
    Column('version', Integer, primary_key=True, autoincrement=False),
    Column('path', String),
    Column('dataqueryname', String),
    Column('username', String, ForeignKey('users.username'), nullable=False),
    Column('dataquerybody', String, nullable=False),
    Column('currentflag', SmallInteger),
    Column('lastmodified', PrismDateTime, default=datetime.datetime.utcnow, onupdate=datetime.datetime.utcnow),
    Column('createdtime', PrismDateTime, default=datetime.datetime.utcnow),
    Column('shared', Boolean, default=False),
    Column('accountid', ForeignKey('accounts.accountid'), nullable=False),
    Index('idx_dataqueries_username', 'username'),
    Index('idx_dataqueries_username_path_currentflag', 'username', 'path', 'currentflag'),
)



from typing import Union

import pytest

from prism._common.const import RankType
from prism._utils.validate_utils import _validate_args
from prism._utils.exceptions import PrismTypeError, PrismValueError


@_validate_args
def factor_backtest(
        self,
        jobname: Union[str, list] = None,
        report: bool = True,
        frequency: str = None,
        bins: int = None,
        rank_method: str = None) -> RankType:
    """
    valid arguments:
        'standard', 'modified', 'dense', 'ordinal', 'fractional'
    """
    return rank_method


def test_periodtype_param():
    assert factor_backtest(None) is None
    assert factor_backtest(
        None, jobname=["job1", "job2"], rank_method="standard", frequency='N', bins=1) == RankType.STANDARD

    # check type
    assert isinstance(factor_backtest(None, "job", rank_method="standard"), RankType)

    # valid periodtype param
    assert factor_backtest(
        None, "job", rank_method="standard") == RankType.STANDARD
    assert factor_backtest(
        None, "job", rank_method="modified") == RankType.MODIFIED
    assert factor_backtest(
        None, "job", rank_method="dense") == RankType.DENSE
    assert factor_backtest(
        None, "job", rank_method="ordinal") == RankType.ORDINAL
    assert factor_backtest(
        None, "job", rank_method="fractional") == RankType.FRACTIONAL


def test_wrong_periodtype_param():
    # wrong value
    with pytest.raises(PrismValueError):
        assert factor_backtest(None, "job1", rank_method='annual')

    with pytest.raises(PrismValueError):
        assert factor_backtest(
            None, "job1", rank_method='Standard') == RankType.STANDARD

    # wrong type
    with pytest.raises(PrismTypeError):
        assert factor_backtest(None, "job", rank_method=1)

    with pytest.raises(PrismTypeError):
        assert factor_backtest(
            None, "job", rank_method='standard', frequency=1) == RankType.STANDARD



from ._auth import login
from ._dataquery import (
    list_dataquery,
    load_dataquery,
    save_dataquery,
    delete_dataquery,
    rename_dataquery,
    extract_dataquery,
    get_data,
    view_data,
)
from ._taskquery import (
    list_taskquery,
    load_taskquery,
    save_taskquery,
    delete_taskquery,
    rename_taskquery,
    extract_taskquery,
)
from ._job import (
    list_job,
    get_job,
    delete_job,
    strategy_backtest_jobs,
    extract_job,
    cancel_job,
    prioritise_job,
)
from ._list import dataitems, packages, _list_dataitem
from ._securitymaster import get_securitymaster, get_securitymaster_advanced
from ._universe import (
    list_universe,
    get_universe,
    upload_universe,
    save_index_as_universe,
    combine_universe,
    delete_universe,
    get_universe_template,
    filter_universe,
    rename_universe,
)

from ._exportdata import (
    list_datafiles,
    retrieve_datafiles,
    rename_datafiles,
    delete_datafiles,
)

from ._portfolio import list_portfolio, save_index_as_portfolio, get_portfolio, rename_portfolio, delete_portfolio
from ._gui import job_manager, preference_setting, securitymaster_search, dataitem_search, finder, open_document



import pytest


@pytest.fixture
def commonly_used_dataitems(prism):
    mc = prism.market.market_cap(currency='trade')
    close = prism.market.close(adjustment='split', currency='trade').resample('D', 21)

    cash = (
        prism.financial.balance_sheet(dataitemid=100004, periodtype='Q', preliminary='keep', currency='trade')
            .fillna(method='ffill', n=4)
            .fillna(0)
            .resample('D', 365)
    )

    debt = (
        prism.financial.balance_sheet(dataitemid=100151, periodtype='Q', preliminary='keep', currency='trade')
            .fillna(method='ffill', n=4)
            .fillna(0)
            .resample('D', 365)
    )

    ev = prism.market.enterprise_value(currency='trade')
    return [mc, close, cash, debt, ev]


# -------------------------------------------------------------------------------------------------------------------- #
#                                                     Value Factors                                                    #
# -------------------------------------------------------------------------------------------------------------------- #

@pytest.fixture
def sales_related_factors(commonly_used_dataitems, prism):
    mc, _, _, _, ev = commonly_used_dataitems
    sales = (
        prism.financial.income_statement(dataitemid=100580, periodtype='LTM', preliminary='keep', currency='trade')
            .fillna(method='ffill', n=4)
            .resample('D', 365)
    )
    sales_fwd = (
        prism.estimate.consensus(dataitemid=200256, periodtype='NTM', periodout=0, currency='trade')
            .resample('D', 365)
    )

    #101. Sales to Price
    s2p = sales / mc

    #102. 5 year Average Sales to Price
    # s2p_5yr = s2p.n_periods_mean(5*365) #daily
    s2p_5yr = s2p.resample('M', 31).n_periods_mean(5*12) #monthly

    #103. Forward Sales to Price
    s2p_fwd = sales_fwd / mc

    #104. Sales to Enterprise Value(EV)
    s2ev = sales / ev

    #105. 5 year average Sales to Enterprise Value
    s2ev_5yr = (sales / ev).resample('BD', 10).n_periods_mean(5*260)
    #s2ev_5yr = s2ev.resample('M', 31).n_period_pct_mean(5*12) #monthly

    #106. Forward Sales to EV
    s2ev_fwd = sales_fwd / ev
    return {'s2p': s2p, 's2p_5yr': s2p_5yr, 's2p_fwd': s2p_fwd, 's2ev': s2ev, 's2ev_5yr': s2ev_5yr, 's2ev_fwd': s2ev_fwd}


@pytest.fixture
def book_related_factors(commonly_used_dataitems, prism):
    mc, _, _, _, _ = commonly_used_dataitems
    book = (
        prism.financial.balance_sheet(dataitemid=100075, periodtype='Q', preliminary='keep', currency='trade')
            .fillna(method='ffill', n=4)
            .resample('D', 365)
    )

    goodwill = (
        prism.financial.balance_sheet(dataitemid=100023, periodtype='Q', preliminary='keep', currency='trade')
            .fillna(method='ffill', n=4)
            .fillna(0)
            .resample('D', 365)
    )

    intangible = (
        prism.financial.balance_sheet(dataitemid=100024, periodtype='Q', preliminary='keep', currency='trade')
            .fillna(method='ffill', n=4)
            .fillna(0)
            .resample('D', 365)
    )

    book_tangible = book - goodwill - intangible

    #107. Book to Price
    b2p = book / mc

    #108. Tangible Book to Price
    b2p_tangible = book_tangible / mc

    #129. 12M Book to Price Z-Score
    b2p_z = b2p.n_periods_z_score(365)

    return {'b2p': b2p, 'b2p_tangible': b2p_tangible, 'b2p_z': b2p_z}


@pytest.fixture
def debt_related_factor(commonly_used_dataitems):
    _, _, cash, debt, ev = commonly_used_dataitems
    d2ev = (debt - cash) / ev
    return {'d2ev': d2ev}


@pytest.fixture
def earnings_related_factors(commonly_used_dataitems, prism):
    mc, _, _, _, ev = commonly_used_dataitems
    earnings = (
        prism.financial.income_statement(dataitemid=100647, periodtype='LTM', preliminary='keep', currency='trade')
            .fillna(method='ffill', n=4)
            .resample('D', 365)
    )

    earnings_fwd = (
        prism.estimate.consensus(dataitemid=200219, periodtype='NTM', periodout=0, currency='trade')
            .resample('D', 365)
    )

    ebidta = (
        prism.financial.income_statement(dataitemid=100723, periodtype='LTM', preliminary='keep', currency='trade')
            .fillna(method='ffill', n=4)
            .resample('D', 365)
    )

    ebidta_fwd = (
        prism.estimate.consensus(dataitemid=200099, periodtype='NTM', periodout=0, currency='trade')
            .resample('D', 365)
    )

    gp = (
        prism.financial.income_statement(dataitemid=100595, periodtype='LTM', preliminary='keep', currency='trade')
            .fillna(method='ffill', n=4)
            .resample('D', 365)
    )

    #110. Earnings to Price
    e2p = earnings / mc

    #111. 5 year Average Earnings to Price
    e2p_5yr = e2p.n_periods_mean(5*365) #daily
    #e2p_5yr = e2p.resample('M', 31).n_periods_mean(5*12) #monthly

    #112. Forward Earnings to Price
    e2p_fwd = earnings_fwd / mc

    #113. EBITDA to EV
    ebidta2ev = ebidta / ev

    #114. Forward BITDA to EV
    ebitda2ev_fwd = ebidta_fwd / ev

    #115. Gross Profit to Price
    gp2p = gp / mc

    #130. 12M Earnings to Price Z-Score
    e2p_z = e2p.n_periods_z_score(365)

    #131. 4w Forward Earnings to Price Change
    e2pchg_fwd_4w = (earnings_fwd / mc).n_periods_pct_change(30)

    #132. 8w Forward Earnings to Price Change
    e2pchg_fwd_8w = (earnings_fwd / mc).n_periods_pct_change(60)

    return {
        'e2p': e2p,
        'e2p_5yr': e2p_5yr,
        'e2p_fwd': e2p_fwd,
        'ebidta2ev': ebidta2ev,
        'ebitda2ev_fwd': ebitda2ev_fwd,
        'gp2p': gp2p,
        'e2p_z': e2p_z,
        'e2pchg_fwd_4w': e2pchg_fwd_4w,
        'e2pchg_fwd_8w': e2pchg_fwd_8w
    }


@pytest.fixture
def cash_flow_related_factors(commonly_used_dataitems, prism):
    mc, _, _, _, ev = commonly_used_dataitems
    cash_op = (
        prism.financial.cash_flow(dataitemid=100412, periodtype='LTM', preliminary='keep', currency='trade')
            .fillna(method='ffill', n=4)
            .resample('D', 365)
    )

    cash_op_fwd = (
        prism.estimate.consensus(dataitemid=200063, periodtype='NTM', periodout=0, currency='trade')
            .resample('D', 365)
    )

    cash_free = (
        prism.financial.cash_flow(dataitemid=100513, periodtype='LTM', preliminary='keep', currency='trade')
            .fillna(method='ffill', n=4)
            .resample('D', 365)
    )

    cash_free_fwd = (
        prism.estimate.consensus(dataitemid=200153, periodtype='NTM', periodout=0, currency='trade')
            .resample('D', 365)
    )

    #117. Operating Cash Flow to Price
    ocf2p = cash_op / mc

    #118. Foward Operating Cash Flow to Price
    ocf2p_fwd = cash_op_fwd / mc

    #119. Free Cash Flow to Price
    fcf2p = cash_free / mc

    #120. Forward Free Cash Flow to Price
    fcf2p_fwd = cash_free_fwd / mc

    #121. Operating Cash Flow to EV
    ocf2ev = cash_op / ev

    #122. Forward Operating Cash Flow to EV
    ocf2ev_fwd = cash_op_fwd / ev

    #123. Operating Cash Flow to EV
    fcf2ev = cash_free / ev

    #124. Forward Operating Cash Flow to EV
    fcf2ev_fwd = cash_free_fwd / ev

    return {
        'ocf2p': ocf2p,
        'ocf2p_fwd': ocf2p_fwd,
        'fcf2p': fcf2p,
        'fcf2p_fwd':fcf2p_fwd,
        'ocf2ev': ocf2ev,
        'ocf2ev_fwd': ocf2ev_fwd,
        'fcf2ev': fcf2ev,
        'fcf2ev_fwd': fcf2ev_fwd
    }


@pytest.fixture
def yield_related_factors(commonly_used_dataitems, prism):
    mc, close, _, _, _ = commonly_used_dataitems
    dps = (
        prism.financial.dps(dataitemid=100549, periodtype='LTM', preliminary='keep', currency='trade')
            .fillna(method='ffill', n=2)
            .fillna(0)
            .resample('D', 365)
    )

    dps_fwd = (
        prism.estimate.consensus(dataitemid=200069, periodtype='NTM', periodout=0, currency='trade')
            .fillna(0)
            .resample('D', 365)
    )

    buyback_comm = (
        prism.financial.cash_flow(dataitemid=100431, periodtype='LTM', preliminary='keep', currency='trade')
            .fillna(method='ffill', n=2).fillna(0) * -1
    )

    buyback_pref = (
        prism.financial.cash_flow(dataitemid=100433, periodtype='LTM', preliminary='keep', currency='trade')
            .fillna(method='ffill', n=2).fillna(0) * -1
    )

    div = (
        prism.financial.cash_flow(dataitemid=100438, periodtype='LTM', preliminary='keep', currency='trade')
            .fillna(method='ffill', n=2).fillna(0) * -1
    )


    #125. Dividend Yield
    dy = dps / close

    #126. Forward Dividend Yield
    dy_fwd = dps_fwd / close

    #127. Buyback Yield
    bby = (buyback_comm + buyback_pref ).resample('D', 365) / mc

    #127. Total Yield
    ty = (buyback_comm + buyback_pref + div).resample('D', 365) / mc

    return {
        'dy': dy,
        'dy_fwd': dy_fwd,
        'bby': bby,
        'ty': ty
    }



# -------------------------------------------------------------------------------------------------------------------- #
#                                                     Growth Factors                                                    #
# -------------------------------------------------------------------------------------------------------------------- #

@pytest.fixture
def growth_sales_related_factor(prism):
    sales = prism.financial.income_statement(dataitemid=100580, periodtype='LTM', preliminary='keep', currency='trade')
    sales_fwd = prism.estimate.consensus(dataitemid=200256, periodtype='NTM', periodout=0, currency='trade').resample('D', 365)
    sales_act = prism.estimate.actual(dataitemid=200032, periodtype='A', currency='trade').resample('D', 365)

    #201. 1 Year Sales Growth
    sg_1yr = sales.n_fiscal_quarters_pct_change(4, positive_denominator=True).fillna(method='ffill', n=4).resample('D', 365)

    #202. 3 Year Sales Growth
    sg_3yr = sales.n_fiscal_quarters_pct_change(12, positive_denominator=True).fillna(method='ffill', n=4).resample('D', 365)

    #204. Forward Sales Growth
    sg_fwd =  (sales_fwd / sales_act) - 1
    return {
        'sg_1yr': sg_1yr,
        'sg_3yr': sg_3yr,
        'sg_fwd': sg_fwd
    }


@pytest.fixture
def growth_debt_related_factor(prism):
    capex = prism.financial.cash_flow(dataitemid=100413, periodtype='LTM', preliminary='keep', currency='trade')

    #205. 1 Year Capital Expenditure Growth
    capex_1yr = capex.n_fiscal_quarters_pct_change(4).fillna(method='ffill', n=4).resample('D', 365)
    return {'capex_1yr': capex_1yr}


@pytest.fixture
def growth_earnings_related_factor(prism):
    earnings = prism.financial.income_statement(dataitemid=100647, periodtype='LTM', preliminary='keep', currency='trade')

    #206. 1 Year Earnings Growth
    eg_1yr = earnings.n_fiscal_quarters_pct_change(4, positive_denominator=True).fillna(method='ffill', n=4).resample('D', 365)

    #207. 3 Year Earnings Growth
    eg_3yr = earnings.n_fiscal_quarters_pct_change(12, positive_denominator=True).fillna(method='ffill', n=4).resample('D', 365)

    #208. 5 Year Earnings Growth
    eg_5yr = earnings.n_fiscal_quarters_pct_change(20, positive_denominator=True).fillna(method='ffill', n=4).resample('D', 365)

    return {'eg_1yr': eg_1yr, 'eg_3yr': eg_3yr, 'eg_5yr': eg_5yr}


@pytest.fixture
def growth_cash_flow_related_factors(prism):
    cash_op = prism.financial.cash_flow(dataitemid=100412, periodtype='LTM', preliminary='keep', currency='trade')
    cash_op_fwd = prism.estimate.consensus(dataitemid=200063, periodtype='NTM', periodout=0, currency='trade').resample('D', 365)
    cash_op_act = prism.estimate.actual(dataitemid=200005, periodtype='A', currency='trade').resample('D', 365)
    cash_free = prism.financial.cash_flow(dataitemid=100513, periodtype='LTM', preliminary='keep', currency='trade')
    cash_free_fwd = prism.estimate.consensus(dataitemid=200153, periodtype='NTM', periodout=0, currency='trade').resample('D', 365)
    cash_free_act = prism.estimate.actual(dataitemid=200005, periodtype='A', currency='trade').resample('D', 365)


    #217. 1 Year Operating Cash Flow Growth
    ocfg_1yr = cash_op.n_fiscal_quarters_pct_change(4, positive_denominator=True).fillna(method='ffill', n=4).resample('D', 365)

    #218. 3 Year Operating Cash Flow Growth
    ocfg_3yr = cash_op.n_fiscal_quarters_pct_change(12, positive_denominator=True).fillna(method='ffill', n=4).resample('D', 365)

    #219. 5 Year Operating Cash Flow Growth
    ocfg_5yr = cash_op.n_fiscal_quarters_pct_change(20, positive_denominator=True).fillna(method='ffill', n=4).resample('D', 365)

    #220. Operating Cash Flow Forward Growth
    ocfg_fwd = (cash_op_fwd / cash_op_act) - 1

    #221. 1 Year Free Cash Flow Growth
    fcfg_1yr = cash_free.n_fiscal_quarters_pct_change(4, positive_denominator=True).fillna(method='ffill', n=4).resample('D', 365)

    #222. 3 Year Free Cash Flow Growth
    fcfg_3yr = cash_free.n_fiscal_quarters_pct_change(12, positive_denominator=True).fillna(method='ffill', n=4).resample('D', 365)

    #223. 5 Year Free Cash Flow Growth
    fcfg_5yr = cash_free.n_fiscal_quarters_pct_change(20, positive_denominator=True).fillna(method='ffill', n=4).resample('D', 365)

    #224. Free Cash Flow Forward Growth
    fcfg_fwd = (cash_free_fwd / cash_free_act) - 1

    return {
        'ocfg_1yr': ocfg_1yr,
        'ocfg_3yr': ocfg_3yr,
        'ocfg_5yr': ocfg_5yr,
        'ocfg_fwd': ocfg_fwd,
        'fcfg_1yr': fcfg_1yr,
        'fcfg_3yr': fcfg_3yr,
        'fcfg_5yr': fcfg_5yr,
        'fcfg_fwd': fcfg_fwd
    }


@pytest.fixture
def growth_yield_related_factors(prism):
    dps = prism.financial.dps(dataitemid=100549, periodtype='LTM', preliminary='keep', currency='trade')
    dps_fwd = prism.estimate.consensus(dataitemid=200069, periodtype='NTM', periodout=0, currency='trade').fillna(0).resample('D', 365)
    dps_act = prism.estimate.actual(dataitemid=200006, periodtype='A', currency='trade').fillna(0).resample('D', 365)

    buyback_comm = prism.financial.cash_flow(dataitemid=100431, periodtype='LTM', preliminary='keep', currency='trade').fillna(0) * -1
    buyback_pref = prism.financial.cash_flow(dataitemid=100433, periodtype='LTM', preliminary='keep', currency='trade').fillna(0) * -1
    div = prism.financial.cash_flow(dataitemid=100438, periodtype='LTM', preliminary='keep', currency='trade').fillna(0) * -1

    #210. 1 Year Dividend Growth
    dg_1yr = dps.n_fiscal_quarters_pct_change(4, positive_denominator=True).fillna(method='ffill', n=2).resample('D', 365)

    #211. 3 Year Dividend Growth
    dg_3yr = dps.n_fiscal_quarters_pct_change(12, positive_denominator=True).fillna(method='ffill', n=2).resample('D', 365)

    #211. 5 Year Dividend Growth
    dg_5yr = dps.n_fiscal_quarters_pct_change(20, positive_denominator=True).fillna(method='ffill', n=2).resample('D', 365)

    #213. Dividend Forward Growth
    dg_fwd = (dps_fwd / dps_act) - 1

    #214. 1 Year Total Yield Growth
    yg_1yr = (buyback_comm + buyback_pref + div).n_fiscal_quarters_pct_change(4, positive_denominator=True).fillna(method='ffill', n=2).resample('D', 365)

    #215. 3 Year Total Yield Growth
    yg_3yr = (buyback_comm + buyback_pref + div).n_fiscal_quarters_pct_change(12, positive_denominator=True).fillna(method='ffill', n=2).resample('D', 365)

    #216. 5 Year Total Yield Growth
    yg_5yr = (buyback_comm + buyback_pref + div).n_fiscal_quarters_pct_change(20, positive_denominator=True).fillna(method='ffill', n=2).resample('D', 365)

    return {
        'dg_1yr': dg_1yr,
        'dg_3yr': dg_3yr,
        'dg_5yr': dg_5yr,
        'dg_fwd': dg_fwd,
        'yg_1yr': yg_1yr,
        'yg_3yr': yg_3yr,
        'yg_5yr': yg_5yr
    }




import pytest
import json
from sqlalchemy import update

from tests.test_util.dq_table import dataqueries


def update_dataquery(
    sales_related_factors,
    book_related_factors,
    debt_related_factor,
    earnings_related_factors,
    cash_flow_related_factors,
    yield_related_factors,
    growth_sales_related_factor,
    growth_debt_related_factor,
    growth_earnings_related_factor,
    growth_cash_flow_related_factors,
    growth_yield_related_factors,
    db_env
):
    engine = db_env

    with engine.connect() as con:
        for dq_name, dq in sales_related_factors.items():
            con.execute(update(dataqueries).where(dataqueries.c.dataqueryname == dq_name).values(dataquerybody = json.dumps(dq._query)))

        for dq_name, dq in book_related_factors.items():
            con.execute(update(dataqueries).where(dataqueries.c.dataqueryname == dq_name).values(dataquerybody = json.dumps(dq._query)))

        for dq_name, dq in debt_related_factor.items():
            con.execute(update(dataqueries).where(dataqueries.c.dataqueryname == dq_name).values(dataquerybody = json.dumps(dq._query)))

        for dq_name, dq in earnings_related_factors.items():
            con.execute(update(dataqueries).where(dataqueries.c.dataqueryname == dq_name).values(dataquerybody = json.dumps(dq._query)))

        for dq_name, dq in cash_flow_related_factors.items():
            con.execute(update(dataqueries).where(dataqueries.c.dataqueryname == dq_name).values(dataquerybody = json.dumps(dq._query)))

        for dq_name, dq in yield_related_factors.items():
            con.execute(update(dataqueries).where(dataqueries.c.dataqueryname == dq_name).values(dataquerybody = json.dumps(dq._query)))

        for dq_name, dq in growth_sales_related_factor.items():
            con.execute(update(dataqueries).where(dataqueries.c.dataqueryname == dq_name).values(dataquerybody = json.dumps(dq._query)))

        for dq_name, dq in growth_debt_related_factor.items():
            con.execute(update(dataqueries).where(dataqueries.c.dataqueryname == dq_name).values(dataquerybody = json.dumps(dq._query)))

        for dq_name, dq in growth_earnings_related_factor.items():
            con.execute(update(dataqueries).where(dataqueries.c.dataqueryname == dq_name).values(dataquerybody = json.dumps(dq._query)))

        for dq_name, dq in growth_cash_flow_related_factors.items():
            con.execute(update(dataqueries).where(dataqueries.c.dataqueryname == dq_name).values(dataquerybody = json.dumps(dq._query)))

        for dq_name, dq in growth_yield_related_factors.items():
            con.execute(update(dataqueries).where(dataqueries.c.dataqueryname == dq_name).values(dataquerybody = json.dumps(dq._query)))


def update_modelquery(
    tcmodel_quries,
    riskmodel_queires,
    db_env
):
    engine = db_env

    with engine.connect() as con:
        for mq_name, mq in tcmodel_quries.items():
            con.execute(update(dataqueries).where(dataqueries.c.dataqueryname == mq_name).values(dataquerybody = json.dumps(mq._query)))

        for rm_name, mq in riskmodel_queires.items():
            con.execute(update(dataqueries).where(dataqueries.c.dataqueryname == rm_name).values(dataquerybody = json.dumps(mq._query)))




from .prismcomponent_utils import _get_params, _req_call, plot_tree, are_periods_exclusive
from .validate_utils import get_sm_attributevalue
from .validate_utils import _validate_args
from .auth_utils import (
    _authentication,
    _create_token,
    _find_file_path,
    TokenDoesNotExistError,
    _get_web_authentication_token,
    _delete_token,
)
from .req_builder_utils import (
    get, post, patch, delete, _process_response,
    _process_fileresponse, _fetch_and_parse
)
from .loader import Loader, download



import requests
import warnings
import webbrowser
from typing import List, Union

import orjson
import pandas as pd

from prism._utils.exceptions import PrismNotFoundError, PrismResponseError, PrismTypeError, PrismValueError, PrismAuthError

from ..._common.config import *
from ..._common import const
from prism import _core
from ..._utils import (
    _validate_args,
    get,
    post,
    patch,
    delete,
    _authentication,
    _process_fileresponse,
    _process_response,
    _fetch_and_parse,
    _get_web_authentication_token,
    Loader,
    download,
    plot_tree,
    get_sm_attributevalue,
    are_periods_exclusive,
)
from .preference import get_quiet_mode
from ..._prismcomponent import prismcomponent as pcmpt
from ._universe import parse_universe_to_universeid


__all__ = [
    "list_dataquery",
    "load_dataquery",
    "save_dataquery",
    "rename_dataquery",
    "delete_dataquery",
    "extract_dataquery",
    "get_data",
    "view_data",
]


@_validate_args
def list_dataquery(search: str = None, tree=False):
    """
    Return the list of all saved data queries available.

    Parameters
    ----------
        search : str, default None
            | Search word for data query name, the search is case-insensitive.
            | If None, the entire list is returned.

        tree : bool, default False
            | If True, the folder structure of the data queries is visualized in a UNIX-style.

    Returns
    -------
        bool or pandas.DataFrame
            | If tree is True, print file system tree of data queries.
            | If tree is False, return usable dataqueries in pandas.DataFrame.
            | Columns:

            - *dataqueryid*
            - *dataqueryname*

    Examples
    --------
        >>> prism.list_dataquery()
        dataqueryid  dataqueryname
        0  	           1	          ROE
        1	           2	          PER
        2	           3	          STD
    """
    dataqueries_df = _fetch_and_parse(URL_DATAQUERIES, "pdq", search)
    if dataqueries_df.empty:
        raise PrismNotFoundError("No Dataqueries found.")
    if not tree:
        return dataqueries_df
    plot_tree(dataqueries_df["dataqueryname"].tolist(), "dataquery")


@_validate_args
def load_dataquery(dataquery: Union[str, int]):
    """
    Load specified data queries.

    Parameters
    ----------
        dataquery : str or int
            Data query name (str) or data query id (int) to load.

    Returns
    -------
        prism._PrismComponent
            Return data query component.

    Examples
    --------
        >>> prism.load_dataquery(7)
        ==== Close
        Query Structure
    """
    headers = _authentication()
    dataqueryid = parse_dataquery_to_dataqueryid(dataquery)
    res = requests.get(url=URL_DATAQUERIES + f"/{dataqueryid}", headers=headers)
    if not res.ok:
        raise PrismResponseError(res.json()["message"])
    query = res.json()["rescontent"]["data"]["dataquerybody"]
    return _dataquery_to_component(query)


def _dataquery_to_component(query: dict):
    componentinfo = const.CategoryComponent[
        (const.CategoryComponent["componentid"]==query["componentid"])
        & (const.CategoryComponent["categoryid"]==query["categoryid"])
    ]
    if query["component_type"] == "functioncomponent":
        component = pcmpt._functioncomponent_builder(
            componentinfo["component_name_repr"].values[0],
            query["component_args"],
            *query["children"],
        )
    else:
        category = getattr(_core, componentinfo["component_category_repr"].values[0])
        component_fn = getattr(category, componentinfo["component_name_repr"].values[0])
        component = component_fn(**query["component_args"])
    return component


@_validate_args
def rename_dataquery(old: Union[str, int], new: str):
    """
    Rename data query. Location of data query within the folder structure can be changed using the method.

    Parameters
    ----------
        old: str or int
            Name of existing data query (str) or data query id (int) to be renamed.
        new: str
            New name of the data query.

    Returns
    -------
        status: dict
            Status of rename_dataquery.

    Examples
    --------
        >>> prism.list_dataquery()
        dataqueryid  dataqueryname
        0  	         1	          ROE
        1	           2	          PER
        2	           3	          STD

        >>> prism.rename_dataquery(old="PER", name="Price to Earnings")
        {
        'status': 'Success',
        'message': 'Dataquery renamed',
        'result': [{'resulttype': 'dataqueryid', 'resultvalue': 2}]
        }

        >>> prism.list_dataquery()
        dataqueryid      dataqueryname
        0  	         1	              ROE
        1	           2  Price to Earnings
        2	           3	              STD
    """
    should_overwrite, err_msg = should_overwrite_dataquery(new, "renaming")
    if not should_overwrite:
        return err_msg
    dataqueryid = parse_dataquery_to_dataqueryid(old)
    ret = patch(URL_DATAQUERIES + f"/{dataqueryid}", {"newpath": new + ".pdq"}, None)
    print(ret["message"])
    return ret


@_validate_args
def save_dataquery(component, dataqueryname: str):
    """
    Save data query.

    Parameters
    ----------
        component : prism._PrismComponent
            New PrismComponent to be saved.

        dataqueryname : str
            Name of the new data query.

    Returns
    -------
        status: dict
            Status of save_dataquery.

    Examples
    --------
        >>> prism.list_dataquery()
        dataqueryid  dataqueryname
        0  	         1	          ROE
        1	           2	          PER
        2	           3	          STD

        >>> c = prism.market.close()
        >>> prism.save_dataquery(component=c, name="close")
        {
        'status': 'Success',
        'message': 'Dataquery saved',
        'result': [{'resulttype': 'dataqueryid', 'resultvalue': 4}]
        }

        >>> prism.list_dataquery()
        dataqueryid  dataqueryname
        0  	         1	          ROE
        1	           2	          PER
        2	           3	          STD
        3	           4	        close
    """
    should_overwrite, err_msg = should_overwrite_dataquery(dataqueryname, "saving")
    if not should_overwrite:
        return err_msg

    ret = post(URL_DATAQUERIES, {"path": dataqueryname + ".pdq"}, component._query)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def delete_dataquery(dataquery: Union[str, int]):
    """
    Delete data query.

    Parameters
    ----------
        dataquery: str or int
            Data query name (str) or data query id (int) to delete.

    Returns
    -------
        status: dict
            Status of delete_dataquery.

    Examples
    --------
        >>> prism.list_dataquery()
        dataqueryid  dataqueryname
        0  	         1	          ROE
        1	           2	          PER
        2	           3	          STD

        >>> prism.delete_dataquery(1)
        {
        'status': 'Success',
        'message': 'Dataquery deleted',
        'result': [{'resulttype': 'dataqueryid', 'resultvalue': 1}]
        }

        >>> prism.list_dataquery()
        dataqueryid  dataqueryname
        0	           2	          PER
        1	           3	          STD

        >>> prism.delete_dataquery("STD")
        {
        'status': 'Success',
        'message': 'Dataquery deleted',
        'result': [{'resulttype': 'dataqueryid', 'resultvalue': 3}]
        }

        >>> prism.list_dataquery()
        dataqueryid  dataqueryname
        0	           2	          PER
    """

    dataqueryid = parse_dataquery_to_dataqueryid(dataquery)
    ret = delete(URL_DATAQUERIES + f"/{dataqueryid}")
    print(f'{ret["message"]}')
    return ret


@_validate_args
def extract_dataquery(component: pcmpt._PrismComponent, return_code=False):
    """
    Generate code which reproduces the data query provided by the component. If return_code is False, the method returns None and prints the code.

    Parameters
    ----------
        component : PrismComponent
            PrismComponent whose data query is to be extracted. Data component or function component should be provided.

        return_code : bool, default False
            If True, the method returns the code. Else the code is printed as system output, returning None.

    Returns
    -------
        str or None
            | If return_code is True, return code in string.
            | If return_code is False, return None.

    Examples
    --------
        >>> ni = prism.financial.income_statement(100639, periodtype='LTM', package=None)
        >>> mcap = prism.market.market_cap(package=None)
        >>> ep = mcap/ni

        >>> prism.extract_dataquery(ep)
        x0 = prism.market.marketcap(currency=None, package="CIQ Market")
        x1 = prism.financial.income_statement(dataitemid=100639, periodtype="LTM", package="CIQ Premium Financials", preliminary=True, currency=None)
        x2 = x0 / x1
    """
    code = post(URL_DATAQUERIES + "/extract", {"dialect": "python"}, component._query)
    if return_code:
        return code
    else:
        print(code)


@_validate_args
def get_data(
    component: Union[pcmpt._AbstractPrismComponent, list],
    universe: Union[str, int] = None,
    startdate: str = None,
    enddate: str = None,
    shownid: list = None,
    display_pit: bool = True,
    name: Union[str, list] = None,
) -> pd.DataFrame:
    """
    Return data corresponding to the query of the component. Multiple components can be passed on, in which case tuple of dataframes are returned.

    Parameters
    ----------
        component : PrismComponent or list
            | PrismComponent which hold the logic to query data.

        universe : str or int, default None
            | Universe name (*str*) or universe id (*int*) used to query data.
            | Some components do not require universe information (eg. Exchange Rate), in which case to be left None.

        startdate : str, default None
            | Start date of the data. The data includes start date.
            | If None, the start date of the universe is used. The same applies when the input is earlier than the universe start date.

        enddate : str, default None
            | End date of the data. The data excludes end date.
            | If None, the end date of the universe is used. The same applies when the input is later than the universe end date.

        shownid : list, default None
            | List of Security Master attributes to display with the data.
            | See prism securitymaster list_attribute for full list of Security Master attributes.
            | If None, default attributes set in preferences is shown.
            | If empty list ([]), no attribute is shown.

        display_pit: bool, default True
            | If True, the data is shown as point-in-time format.
            | If False, the data is collapsed to display from the most recent view point.

        name : str or list, default None
            | Column names of the data to display.
            | If one component is passed to the function, accepts either string or list with one element.
            | If multiple components is passed to the function, accepts list of string.
            | If None:

            - If data component is passed, the column name is implicitly decided following the name of the data component.
            - If function component is passed, the default column name is 'value'.


    Returns
    -------
        pandas.DataFrame
            | Requested data component's DataFrame
            | Columns:

            - *listingid*
            - *date*
            - *value: can be provided by the user via 'name'*
            - *<attributes>: provided by the user via 'shownid'*
            - *period (Optional): for financial datacomponents*

    Examples
    --------
        >>> close = prism.market.close(adjustment='All', package='Prism Market')
        >>> close_df = prism.get_data(component=[close], universe='KOSPI 200 Index', startdate='2015-01-01', enddate='2020-12-31', shownid=['ticker'])
        >>> close_df
                listingid        date          Close   ticker
        0        20108718  2015-01-02    7740.611273  A004430
        1        20108718  2015-01-05    7874.399616  A004430
        2        20108718  2015-01-06    7903.068547  A004430
        3        20108718  2015-01-07    8313.989886  A004430
        4        20108718  2015-01-08    8161.088923  A004430
        ...           ...         ...            ...      ...
        298992  686744025  2020-12-23  151197.660000  A352820
    """

    query = []
    cmpts = set()
    if isinstance(name, list) & (name is not None):
        if any([not isinstance(n, str) for n in name]):
            raise PrismTypeError('Names shoud be string')

    def add_cmpts(o):
        cmpts = set()
        if o["component_type"] in ["datacomponent", "modelcomponent"]:
            cmpts.add(o["componentid"])
        else:
            for c in o["children"]:
                cmpts = cmpts | add_cmpts(c)
        return cmpts

    if not isinstance(component, list):
        component = [component]
    for o in component:
        if isinstance(o, pcmpt._AbstractPrismComponent):
            query.append(o._query)
            cmpts = add_cmpts(o._query)
        else:
            raise PrismTypeError(f"Please provide Components into get data")

    if all(
        (~const.DataComponents[const.DataComponents["componentid"].isin(cmpts)]["need_universe"]).tolist()
        + [False for c in cmpts if c in const.CategoryComponent[const.CategoryComponent["componenttype"]=="model"]["componentid"].tolist()]
    ):
        universeid = None
    else:
        universeid, _ = parse_universe_to_universeid(universe)

        universe_info = get(f"{URL_UNIVERSES}/{universeid}/info")
        universe_startdate = universe_info["Start Date"].values[0]
        universe_enddate = universe_info["End Date"].values[0]

        universe_period_violated = are_periods_exclusive(universe_startdate, universe_enddate, startdate, enddate)

        if universe_period_violated:
            raise PrismValueError(
                f'Query period should overlap with universe period ({str(universe_startdate).split("T")[0]} ~ {str(universe_enddate).split("T")[0]})'
            )

    default_shownid = True
    if (shownid is not None) and (len(shownid) == 0):
        shownid = None
        default_shownid = False
    if shownid is not None:
        shownid = [get_sm_attributevalue(a) for a in shownid]
    component_names = set([c._query["component_name"] for c in component])
    if const.FunctionComponents is None:
        raise PrismAuthError("Please Login First")
    aggregatecomponents = set(const.FunctionComponents[const.FunctionComponents["is_aggregate"]]["component_name_repr"].unique())
    if (len(component_names - aggregatecomponents) == 0) & (shownid is not None):
        warnings.warn(f"Shownid will be ignored for: {list(component_names & aggregatecomponents)}")

    params = {
        "universeid": universeid,
        "startdate": startdate,
        "enddate": enddate,
        "shownid": shownid,
        "default_shownid": default_shownid,
        "datanames": name,
        "return_url": False,
        "display_pit": display_pit,
    }
    headers = _authentication()
    with Loader("\rFetching Data...") as l:
        try:
            res = requests.post(url=URL_DATAQUERIES + "/query", params=params, json=query, headers=headers)
        except:
            l.stop()
            raise PrismResponseError("Request ended with an error!")
        if res.status_code >= 400:
            l.stop()
            return _process_fileresponse(res, "data")
    links = _process_response(res)["urls"]
    res = requests.get(url=links, stream=True)
    ret = download(res)
    dfs, metadata = _process_fileresponse(res, ret)
    order = metadata.get("order")
    ret = []
    for idx, i in enumerate(order):
        if isinstance(component[idx], _core._model.riskmodel._PrismRiskModelComponent):
            if len(dfs) == 0: break
            data_i = {}
            date = metadata["uniqueids"][i].get("date")
            listingid = metadata["uniqueids"][i].get("Listing ID")
            attribute = metadata["uniqueids"][i].get("attribute")
            drop = metadata["uniqueids"][i].get("drop")
            for d in date:
                value = dfs[f"{i}/{d}.parquet"]
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    value.columns = listingid[d]
                    value.index = listingid[d]
                    # value.attrs["attributes"] = attribute[d] if shownid is not None else None
                    # value.attrs["dropped"] = drop
                    # def attributes(self): return self.attrs["attributes"]
                    # value.attributes = attributes.__get__(value)
                    # def dropped(self): return self.attrs["dropped"]
                    # value.dropped = dropped.__get__(value)
                    value.attributes = attribute[d] if shownid is not None else None
                    value.dropped = drop[d]
                data_i[d.split(" ")[0]] = value
        else:
            data_i = dfs[f"{i}.parquet"]
        ret.append(data_i)
    return ret[0] if len(ret) == 1 else ret


@_validate_args
def view_data(
    component: Union[pcmpt._AbstractPrismComponent, list],
    universe: Union[str, int] = None,
    startdate: str = None,
    enddate: str = None,
    shownid: list = None,
    name: Union[str, list] = None,
):
    """
    Open data viewer window on a new browser tab. The data corresponding to the query of the component gets displayed on the data viewer. Multiple components can be passed on, in which case a merged table is displayed.

    Parameters
    ----------
        component : PrismComponent or list
            PrismComponent which hold the logic to query data.

        universe : str or int, default None
            | Universe name (str) or universe id (int) used to query data.
            | Some components do not require universe information (eg. Exchange Rate), in which case to be left None.

        startdate : str, default None
            | Start date of the data. The data includes start date.
            | If None, the start date of the universe is used. The same applies when the input is earlier than the universe start date.

        enddate : str, default None
            | End date of the data. The data excludes end date.
            | If None, the end date of the universe is used. The same applies when the input is later than the universe end date.

        shownid : list, default None
            | List of Security Master attributes to display with the data.
            | See 'prism securitymaster list_attribute' for full list of Security Master attributes.
            | If None, default attributes set in preferences is shown.
            | If empty list ([]), no attribute is shown.

        name : str or list, default None
            | Column names of the data to display.
            | If one component is passed to the function, accepts either string or list with one element.
            | If multiple components is passed to the function, accepts list of string.
            | If None:

            - If data component is passed, the column name is implicitly decided following the name of the data component.
            - If function component is passed, the default column name is 'value'.

        include_na : bool, default False
            | If True, universe data is left-joined to show missing (NA) values.
            | If False, the NA values are dropped.

    Returns
    -------
        None

    Notes
    -----
        The user will be asked to login if not already logged in.

    Examples
    --------
        >>> close = prism.market.close()
        >>> close.view_data(
                universe='Korea Stock Price 200 Index',
                startdate='2010-01-01',
                enddate='2011-01-01',
                shownid=['SEDOL', 'Country', 'GICS Group', 'GVKEY', 'Companyname', 'Fitch Issuer ID', 'ISIN']
            )
    """
    query = []
    cmpts = set()
    if isinstance(name, list) & (name is not None):
        if any([not isinstance(n, str) for n in name]):
            raise PrismTypeError('Names shoud be string')

    def add_cmpts(o):
        cmpts = set()
        if o["component_type"] == "datacomponent":
            cmpts.add(o["componentid"])
        else:
            for c in o["children"]:
                cmpts = cmpts | add_cmpts(c)
        return cmpts

    if not isinstance(component, list):
        component = [component]
    for o in component:
        if isinstance(o, pcmpt._AbstractPrismComponent):
            query.append(o._query)
            cmpts = add_cmpts(o._query)
        else:
            raise PrismTypeError(f"Please provide Components into get data.")

    if all((~const.DataComponents[const.DataComponents["componentid"].isin(cmpts)]["need_universe"]).tolist()):
        universeid = None
    else:
        universeid, universename = parse_universe_to_universeid(universe)

    universe_info = get(f"{URL_UNIVERSES}/{universeid}/info")
    universe_startdate = universe_info["Start Date"].values[0]
    universe_enddate = universe_info["End Date"].values[0]

    universe_period_violated = are_periods_exclusive(universe_startdate, universe_enddate, startdate, enddate)

    if universe_period_violated:
        raise PrismValueError(
            f'Query period should overlap with universe period ({str(universe_startdate).split("T")[0]} ~ {str(universe_enddate).split("T")[0]})'
        )

    if shownid is not None:
        shownid = [get_sm_attributevalue(a) for a in shownid]
    params = {
        "universeid": universeid,
        "startdate": startdate,
        "enddate": enddate,
        "shownid": shownid,
        "view": True,
        "return_url": True,
        "datanames": name,
    }
    headers = _authentication()
    with Loader("\rFetching Data...") as l:
        try:
            res = requests.post(url=URL_DATAQUERIES + "/query", params=params, json=query, headers=headers)
        except:
            l.stop()
            raise PrismResponseError("Request ended with an error!")
        if res.status_code >= 400:
            l.stop()
            return _process_response(res)
    links = _process_response(res)["urls"]
    if len(links) == 0:
        print("The dataframe is empty.")
        return
    else:
        links = links[0]
    view_data_link = ROOT_EXT_WEB_URL + "/view?"
    for k, v in links.items():
        view_data_link += f"{k}={v}&"
    view_data_link += f'universe={universename.replace("&", "%26")}&startdate={startdate}&enddate={enddate}&'
    web_auth_token = _get_web_authentication_token()
    view_data_link += f"token={web_auth_token}"
    webbrowser.open(view_data_link, new=2)


@_validate_args
def parse_dataquery_to_dataqueryid(dataquery: Union[str, int]):
    dataqueries_df = _fetch_and_parse(URL_DATAQUERIES, "pdq")
    if dataqueries_df.empty:
        raise PrismNotFoundError("No dataqueries Found.")

    if isinstance(dataquery, int):
        dataqueries_df = dataqueries_df[dataqueries_df["dataqueryid"] == dataquery]
    elif isinstance(dataquery, str):
        dataqueries_df = dataqueries_df[dataqueries_df["dataqueryname"] == dataquery]
    else:
        raise PrismValueError("Please provide dataquery name or dataquery id for dataquery.")

    if dataqueries_df.empty:
        raise PrismNotFoundError(f"No dataquery matching: {dataquery}")

    dataqueryid = dataqueries_df["dataqueryid"].values[0]
    return dataqueryid


def should_overwrite_dataquery(dataqueryname, operation):
    if get_quiet_mode():
        return True, None
    dataqueries_df = _fetch_and_parse(URL_DATAQUERIES, "pdq")
    if dataqueries_df.empty:
        return True, None

    if dataqueryname in dataqueries_df["dataqueryname"].to_list():
        overwrite = input(f"{dataqueryname} already exists. Do you want to overwrite? (Y/N) \n")
        if overwrite.lower() == "y":
            return True, None
        elif overwrite.lower() == "n":
            return False, f"Abort {operation} dataquery."
        else:
            return False, "Please provide a valid input."
    else:
        return True, None



import pytest

from prism._utils.validate_utils import _validate_args
from prism._utils.exceptions import PrismTypeError, PrismValueError


@_validate_args
def get_portfolio_of_startdate(portfolio,
                               startdate: str = None):
    return startdate


def test_startdate_param():
    # Valid Date
    assert get_portfolio_of_startdate(
        'S&P 500', startdate='2023-04-05') == '2023-04-05'
    assert get_portfolio_of_startdate(
        'S&P 500', startdate='2023-4-5') == '2023-4-5'
    assert get_portfolio_of_startdate(
        'S&P 500', startdate='2023-4') == '2023-4'
    assert get_portfolio_of_startdate(
        'S&P 500', startdate='2023') == '2023'
    assert get_portfolio_of_startdate('S&P 500', '2023.01.11') == '2023.01.11'

    # check type
    assert isinstance(get_portfolio_of_startdate('S&P 500', startdate='2023-04-05'), str)

    # BEGIN DATE
    assert get_portfolio_of_startdate('S&P 500', '1700-01-01') == '1700-01-01'

    # ACTIVE DATE
    assert get_portfolio_of_startdate('S&P 500', '2199-12-31') == '2199-12-31'

    # Looks wrong but valid case. (ex) '23-12-99' is 2099-12-23
    assert get_portfolio_of_startdate('S&P 500', '23-12-99') == '23-12-99'
    assert get_portfolio_of_startdate('S&P 500', '23.12.99') == '23.12.99'
    assert get_portfolio_of_startdate('S&P 500', '23 12 99') == '23 12 99'
    assert get_portfolio_of_startdate('S&P 500', '23/12/99') == '23/12/99'


def test_wrong_startdate_param():
    # out of valid date
    with pytest.raises(AssertionError):
        assert get_portfolio_of_startdate(
            'S&P 500', '2200-01-01') == '2200-01-01'

    with pytest.raises(AssertionError):
        assert get_portfolio_of_startdate(
            'S&P 500', '1699-12-31') == '1699-12-31'

    # pandas - out of bound
    with pytest.raises(PrismValueError):
        assert get_portfolio_of_startdate('S&P 500', '1600-12-01')

    with pytest.raises(PrismValueError):
        assert get_portfolio_of_startdate('S&P 500', '2300-01-01')

    # parsing error
    with pytest.raises(PrismValueError):
        assert get_portfolio_of_startdate(
            'S&P 500', '2023-12-32') == '2023-12-32'

    with pytest.raises(PrismValueError):
        assert get_portfolio_of_startdate(
            'S&P 500', '2023년 12월 20일') == '2023년 12월 20일'

    with pytest.raises(PrismValueError):
        assert get_portfolio_of_startdate(
            'S&P 500', '1699-12-31typo') == '1699-12-31typo'

    # wrong type
    with pytest.raises(PrismTypeError):
        assert get_portfolio_of_startdate('S&P 500', startdate=20220101)


@_validate_args
def get_portfolio_of_enddate(portfolio,
                             enddate: str = None):
    return enddate


def test_enddate_param():
    # Valid Date
    assert get_portfolio_of_enddate(
        'S&P 500', enddate='2023-04-05') == '2023-04-05'
    assert get_portfolio_of_enddate(
        'S&P 500', enddate='2023-4-5') == '2023-4-5'
    assert get_portfolio_of_enddate(
        'S&P 500', enddate='2023-4') == '2023-4'
    assert get_portfolio_of_enddate(
        'S&P 500', enddate='2023') == '2023'
    assert get_portfolio_of_enddate('S&P 500', '2023.01.11') == '2023.01.11'

    # check type
    assert isinstance(get_portfolio_of_enddate('S&P 500', enddate='2023-04-05'), str)

    # BEGIN DATE
    assert get_portfolio_of_enddate('S&P 500', '1700-01-01') == '1700-01-01'

    # ACTIVE DATE
    assert get_portfolio_of_enddate('S&P 500', '2199-12-31') == '2199-12-31'

    # Looks wrong but valid case. (ex) '23-12-99' is 2099-12-23
    assert get_portfolio_of_enddate('S&P 500', '23-12-99') == '23-12-99'
    assert get_portfolio_of_enddate('S&P 500', '23.12.99') == '23.12.99'
    assert get_portfolio_of_enddate('S&P 500', '23 12 99') == '23 12 99'
    assert get_portfolio_of_enddate('S&P 500', '23/12/99') == '23/12/99'


def test_wrong_enddate_param():
    # out of valid date
    with pytest.raises(AssertionError):
        assert get_portfolio_of_enddate(
            'S&P 500', '2200-01-01') == '2200-01-01'

    with pytest.raises(AssertionError):
        assert get_portfolio_of_enddate(
            'S&P 500', '1699-12-31') == '1699-12-31'

    # pandas - out of bound
    with pytest.raises(PrismValueError):
        assert get_portfolio_of_enddate('S&P 500', '1600-12-01')

    with pytest.raises(PrismValueError):
        assert get_portfolio_of_enddate('S&P 500', '2300-01-01')

    # parsing error
    with pytest.raises(PrismValueError):
        assert get_portfolio_of_enddate(
            'S&P 500', '2023-12-32') == '2023-12-32'

    with pytest.raises(PrismValueError):
        assert get_portfolio_of_enddate(
            'S&P 500', '2023년 12월 20일') == '2023년 12월 20일'

    with pytest.raises(PrismValueError):
        assert get_portfolio_of_enddate(
            'S&P 500', '1699-12-31typo') == '1699-12-31typo'

    # wrong type
    with pytest.raises(PrismTypeError):
        assert get_portfolio_of_enddate('S&P 500', enddate=20220101)


@_validate_args
def get_portfolio(portfolio,
                  startdate: str = None,
                  enddate: str = None):
    return startdate, enddate


def test_get_portfolio():
    assert get_portfolio(
        'S&P 500', startdate='2022-04-05', enddate='2023-04-05') == ('2022-04-05', '2023-04-05')

    assert get_portfolio('S&P 500', '1700-01-01',
                         '2199-12-31') == ('1700-01-01', '2199-12-31')


def test_get_portfolio_with_wrong_param():
    # out of valid date
    with pytest.raises(AssertionError):
        assert get_portfolio(
            'S&P 500', startdate='2023-01-01', enddate='2200-01-01') == ('2200-01-01', '2023-01-01')

    with pytest.raises(AssertionError):
        assert get_portfolio(
            'S&P 500', startdate='1699-12-31', enddate='2030-02-02') == ('1699-12-31', '2030-02-02')

    # pandas - out of bound
    with pytest.raises(PrismValueError):
        assert get_portfolio('S&P 500', startdate='1600-12-01',
                             enddate='2023-04-05') == ('1600-12-01', '2023-04-05')

    with pytest.raises(PrismValueError):
        assert get_portfolio('S&P 500', startdate='2023-04-05',
                             enddate='9999-01-01') == ('2023-04-05', '9999-01-01')

    with pytest.raises(PrismValueError):
        assert get_portfolio('S&P 500', startdate='1600-12-01',
                             enddate='9999-01-01') == ('1600-12-01', '9999-01-01')



import pandas as pd

from ..._common.config import *
from ..._common.const import *
from ..._prismcomponent.prismcomponent import _PrismComponent, _PrismDataComponent
from ..._utils.validate_utils import _validate_args, get_sm_attributevalue
from ..._utils import get


__all__ = ['attribute', 'list_attribute']


_data_category = __name__.split(".")[-1]


class attribute(_PrismDataComponent, _PrismComponent):
    """
    Return the security master attribute for a given listingids. Default frequency is daily.

    Parameters
    ----------
        attribute : str, {"Trading Item ID", "Security ID", "Company ID", "Trade Currency", "MIC", "Ticker", "CIQ Primary", "Country", "Company Name", "GVKEY", "GVKEYIID", "ISIN", "SEDOL", "VALOR", "WKN", "Share Class FIGI", "CUSIP", "CINS", "Barra ID", "FIGI", "Composite FIGI", "SIC", "NAICS", "IBES Ticker", "Compustat Primary", "GICS Sector", "GICS Group", "GICS Industry", "GICS Sub-Industry", "LEI", "SNL Institution ID", "Moody's Issuer Number", "RatingsXpress Entity ID", "MarkIt Red Code", "Fitch Issuer ID", "CMA Entity ID", "Factset Listing ID", "Factset Security ID", "Fackset Company ID", "Factset Entity ID", “Current Ticker”, “ISIN”}
            | Security master attribute includes security codes and meta data from multiple vendors.

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> gics = prism.securitymaster.attribute(attribute="Trade Currency")
        >>> gics.get_data(universe="S&P 500", startdate="2010-01-01", enddate="2015-01-01")
                listingid        date  Trade Currency  Company Name
        0         2586086  2010-01-01             USD     AFLAC INC
        1         2586086  2010-01-02             USD     AFLAC INC
        2         2586086  2010-01-03             USD     AFLAC INC
        3         2586086  2010-01-04             USD     AFLAC INC
        4         2586086  2010-01-05             USD     AFLAC INC
        ...           ...         ...             ...           ...
        914740  344286611  2011-10-27             USD      ITT CORP
        914741  344286611  2011-10-28             USD      ITT CORP
        914742  344286611  2011-10-29             USD      ITT CORP
        914743  344286611  2011-10-30             USD      ITT CORP
        914744  344286611  2011-10-31             USD      ITT CORP
    """

    _component_category_repr = _data_category

    @_validate_args
    def __init__(self, attribute: str, offset: int = 0):
        attributevalue = get_sm_attributevalue(attribute)
        super().__init__(attribute=attributevalue, offset=offset)


def list_attribute():
    """
    Return the all usable attributes in the security master.

    Returns
    -------
        list : Usable attribute.

    Examples
    --------
        >>> prism.securitymaster.list_attribute()
        ['Trading Item ID',
        'VALOR',
        'GVKEY',
        'MarkIt Red Code',
        'Country',
        'GICS Sector',
        'WKN',
        'CINS',
        "Moody's Issuer Number",
        'GICS Group',
        'SEDOL',
        'Company Name',
        'CMA Entity ID',
        'CIQ Primary',
        'NAICS',
        'Factset Security ID',
        'Composite FIGI',
        'Share Class FIGI',
        'FIGI',
        'Compustat Primary',
        'Factset Company ID',
        'Ticker',
        'Factset Entity ID',
        'GVKEYIID',
        'Company ID',
        'Factset Listing ID',
        'LEI',
        'IBES Ticker',
        'CUSIP',
        'GICS Sub-Industry',
        'Barra ID',
        'ISIN',
        'MIC',
        'ISIN',
        'SIC',
        'Security ID',
        'Trade Currency',
        'GICS Industry',
        'Fitch Issuer ID',
        'RatingsXpress Entity ID',
        'Current Ticker',
        'SNL Institution ID']
    """
    smattributes = get(f'{URL_SM}/attributes')
    smattributes_df = pd.DataFrame(smattributes)
    return smattributes_df.sort_values(by='attributeorder')['attributerepr'].tolist()



import pytest


@pytest.fixture
def tcmodel_quries(prism):
    # ------------ almgren ------------
    ret = prism.market.close().n_periods_pct_change(1)
    trade = (ret > 0).map({True: 1000000, False: -1000000})
    trade_amount = trade.abs()
    almgren = prism.tcmodel.almgren(trade_amount, 10, 5)

    # ---------- bidaskspred ----------
    bas = prism.tcmodel.bidaskspread()

    return {'almgren': almgren, 'bidaskspread': bas}


@pytest.fixture
def riskmodel_queires(prism):
    qis = prism.riskmodel.qis("M", 252, False)
    return {'qis': qis}


import pandas as pd

from .._req_builder import _list_dataitem, _dataquery
from ..._common.const import (
    AdjustmentType as _AdjustmentType,
    DilutionType as _DilutionType,
    CurrencyTypeWithReportTrade as _CurrencyTypeWithReportTrade
)
from ..._prismcomponent.prismcomponent import _PrismComponent, _PrismDataComponent
from ..._utils import _get_params, _validate_args, _req_call


__all__ = [
    'open',
    'close',
    'high',
    'low',
    'bid',
    'ask',
    'vwap',
    'market_cap',
    'volume',
    'dividend',
    'exchange_rate',
    'short_interest',
    'split',
    'split_adjustment_factor',
    'dividend_adjustment_factor',
    'shares_outstanding',
    'enterprise_value',
    'implied_market_cap',
    'dataitems',
]


_data_category = __name__.split(".")[-1]


class _PrismMarketComponent(_PrismDataComponent, _PrismComponent):
    _component_category_repr = _data_category


class open(_PrismMarketComponent):
    """
    | Daily open pricing history for equity securities.
    | Default frequency is business daily.

    Parameters
    ----------
        adjustment : str {'all', 'split', 'dividend'}, default 'all'
            | Adjustment for pricing data.

            - all : apply both split and dividend adjustment
            - split : apply only split adjustment
            - dividend : apply only dividend adjustment

        currency : str {'trade', 'report', ISO3 currency}, default 'trade'
            | Desired currency for the pricing data.

            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

        package : str {'Prism Market', 'Compustat', 'CIQ Market', 'MI Integrated Market'}, default 'Prism Market'
            | Desired data package in where the pricing data outputs from.

            .. admonition:: Warning
                :class: warning

                If an invalid package is entered without a license, an error will be generated as output.

    Returns
    -------
        prism._PrismComponent: Historical time-series of daily open prices for stocks

    Examples
    --------
        >>> open_prc = prism.market.open(adjustment='all', package='Prism Market')
        >>> open_df = prism.get_data([open_prc], 'KOSPI 200 Index', startdate='2015-01-01', enddate='2020-12-31', shownid=['ticker'])
        >>> open_df
                listingid        date           Open   ticker
        0        20108718  2015-01-02    7711.942343  A004430
        1        20108718  2015-01-05    7625.935551  A004430
        2        20108718  2015-01-06    7750.167583  A004430
        3        20108718  2015-01-07    7845.730686  A004430
        4        20108718  2015-01-08    8342.658817  A004430
        ...           ...         ...            ...      ...
        298991  686744025  2020-12-23  152679.990000  A352820
        298992  686744025  2020-12-24  151197.660000  A352820
        298993  686744025  2020-12-28  156632.870000  A352820
        298994  686744025  2020-12-29  154656.430000  A352820
        298995  686744025  2020-12-30  156632.870000  A352820
    """
    @_validate_args
    def __init__(
        self,
        adjustment: _AdjustmentType = 'all',
        currency: _CurrencyTypeWithReportTrade = 'trade',
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))


class close(_PrismMarketComponent):
    """
    | Daily close pricing history for equity securities.
    | Default frequency is business daily.

    Parameters
    ----------
        adjustment : str {'all', 'split', 'dividend'}, default 'all'
            | Adjustment for pricing data.

            - all : apply both split and dividend adjustment
            - split : apply only split adjustment
            - dividend : apply only dividend adjustment

        currency : str {'trade', 'report', ISO3 currency}, default 'trade'
            | Desired currency for the pricing data.

            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

        package : str {'Prism Market', 'Compustat', 'CIQ Market', 'MI Integrated Market'}, default 'Prism Market'
            | Desired data package in where the pricing data outputs from.

            .. admonition:: Warning
                :class: warning

                If an invalid package is entered without a license, an error will be generated as output.

    Returns
    -------
        prism._PrismComponent: Historical time-series of daily close prices for stocks

    Examples
    --------
        Obtain daily closing prices for a specific security:

        >>> close = prism.market.close(adjustment='all', package='Prism Market')
        >>> close_df = prism.get_data([close], 'KOSPI 200 Index', startdate='2015-01-01', enddate='2020-12-31', shownid=['ticker'])
        >>> close_df
                listingid        date          Close   ticker
        0        20108718  2015-01-02    7740.611273  A004430
        1        20108718  2015-01-05    7874.399616  A004430
        2        20108718  2015-01-06    7903.068547  A004430
        3        20108718  2015-01-07    8313.989886  A004430
        4        20108718  2015-01-08    8161.088923  A004430
        ...           ...         ...            ...      ...
        298992  686744025  2020-12-23  151197.660000  A352820
        298993  686744025  2020-12-24  156138.760000  A352820
        298994  686744025  2020-12-28  154656.430000  A352820
        298995  686744025  2020-12-29  156632.870000  A352820
        298996  686744025  2020-12-30  158115.200000  A352820
    """
    @_validate_args
    def __init__(
        self,
        adjustment: _AdjustmentType = 'all',
        currency: _CurrencyTypeWithReportTrade = 'trade',
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))


class high(_PrismMarketComponent):
    """
    | Daily high pricing history for equity securities.
    | Default frequency is business daily.

    Parameters
    ----------
        adjustment : str {'all', 'split', 'dividend'}, default 'all'
            | Adjustment for pricing data.

            - all: apply both split and dividend adjustment
            - split: apply only split adjustment
            - dividend: apply only dividend adjustment

        currency : str {'trade', 'report', ISO3 currency}, default 'trade'
            | Desired currency for the pricing data.

            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

        package : str {'Prism Market', 'Compustat', 'CIQ Market', 'MI Integrated Market'}, default 'Prism Market'
            | Desired data package in where the pricing data outputs from.

            .. admonition:: Warning
                :class: warning

                If an invalid package is entered without a license, an error will be generated as output.

    Returns
    -------
        prism._PrismComponent: Historical time-series of daily high prices for stocks

    Examples
    --------
        >>> high = prism.market.high(adjustment='all', package='Prism Market')
        >>> high_df = prism.get_data([high], 'KOSPI 200 Index', startdate='2015-01-01', enddate='2020-12-31', shownid=['ticker'])
        >>> high_df
                listingid        date           High   ticker
        0        20108718  2015-01-02    7855.286996  A004430
        1        20108718  2015-01-05    7893.512237  A004430
        2        20108718  2015-01-06    7931.737478  A004430
        3        20108718  2015-01-07    8323.546196  A004430
        4        20108718  2015-01-08    8352.215127  A004430
        ...           ...         ...            ...      ...
        298991  686744025  2020-12-23  156138.760000  A352820
        298992  686744025  2020-12-24  156632.870000  A352820
        298993  686744025  2020-12-28  157126.980000  A352820
        298994  686744025  2020-12-29  158609.310000  A352820
        298995  686744025  2020-12-30  159103.420000  A352820
    """
    @_validate_args
    def __init__(
        self,
        adjustment: _AdjustmentType = 'all',
        currency: _CurrencyTypeWithReportTrade = 'trade',
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))


class low(_PrismMarketComponent):
    """
    | Daily low pricing history for equity securities.
    | Default frequency is business daily.

    Parameters
    ----------
        adjustment : str, {'all', 'split', 'dividend'}, default 'all'
            | Adjustment for pricing data.

            - all : apply both split and dividend adjustment
            - split : apply only split adjustment
            - dividend : apply only dividend adjustment
        currency : str, {'trade', 'report', ISO3 currency}, default 'trade'
            | Desired currency for the pricing data.

            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)
        package : str, {'Prism Market', 'Compustat', 'CIQ Market', 'MI Integrated Market'}, default 'Prism Market'
            | Desired data package in where the pricing data outputs from.

            .. admonition:: Warning
                :class: warning

                If an invalid package is entered without a license, an error will be generated as output.

    Returns
    -------
        prism._PrismComponent: Historical time-series of daily low prices for stocks

    Examples
    --------
        >>> low = prism.market.low(adjustment='all', package='Prism Market')
        >>> low_df = prism.get_data([low], 'KOSPI 200 Index', startdate='2015-01-01', enddate='2020-12-31', shownid=['ticker'])
        >>> low_df
                listingid       date            Low   ticker
        0        20108718 2015-01-02    7501.703518  A004430
        1        20108718 2015-01-05    7539.928759  A004430
        2        20108718 2015-01-06    7750.167583  A004430
        3        20108718 2015-01-07    7836.174375  A004430
        4        20108718 2015-01-08    8141.976302  A004430
        ...           ...        ...            ...      ...
        298991  686744025 2020-12-23  150703.550000  A352820
        298992  686744025 2020-12-24  149715.330000  A352820
        298993  686744025 2020-12-28  153174.100000  A352820
        298994  686744025 2020-12-29  153174.100000  A352820
        298995  686744025 2020-12-30  155644.650000  A352820
    """
    @_validate_args
    def __init__(
        self,
        adjustment: _AdjustmentType = 'all',
        currency: _CurrencyTypeWithReportTrade = 'trade',
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))


class ask(_PrismMarketComponent):
    """
    | End of day ask pricing history for equity securities.
    | Default frequency is business daily.

    Parameters
    ----------
        adjustment : str, {'all', 'split', 'dividend'}, default 'all'
            | Adjustment for pricing data.

            - all : apply both split and dividend adjustment
            - split : apply only split adjustment
            - dividend : apply only dividend adjustment
        currency : str, {'trade', 'report', ISO3 currency}, default 'trade'
            | Desired currency for the pricing data.

            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)
        package : str, {'Prism Market', 'CIQ Market', 'MI Integrated Market'}, default 'Prism Market'
            | Desired data package in where the pricing data outputs from.

            .. admonition:: Warning
                :class: warning

                If an invalid package is entered without a license, an error will be generated as output.

    Returns
    -------
        prism._PrismComponent: Historical time-series of daily end of day ask prices for stocks

    Examples
    --------
        >>> ask = prism.market.ask(adjustment='all', package='Prism Market')
        >>> ask_df = prism.get_data([ask], 'KOSPI 200 Index', startdate='2015-01-01', enddate='2020-12-31', shownid=['ticker'])
        >>> ask_df
                listingid        date            Ask   ticker
        0        20108718  2015-01-02    7740.611273  A004430
        1        20108718  2015-01-05    7874.399616  A004430
        2        20108718  2015-01-06    7903.068547  A004430
        3        20108718  2015-01-07    8313.989886  A004430
        4        20108718  2015-01-08    8199.314164  A004430
        ...           ...         ...            ...      ...
        298906  686744025  2020-12-23  151691.770000  A352820
        298907  686744025  2020-12-24  156138.760000  A352820
        298908  686744025  2020-12-28  155150.540000  A352820
        298909  686744025  2020-12-29  157126.980000  A352820
        298910  686744025  2020-12-30  158115.200000  A352820
    """
    @_validate_args
    def __init__(
        self,
        adjustment: _AdjustmentType = 'all',
        currency: _CurrencyTypeWithReportTrade = 'trade',
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))


class bid(_PrismMarketComponent):
    """
    | End of day bid pricing history for equity securities.
    | Default frequency is business daily.

    Parameters
    ----------
        adjustment : str, {'all', 'split', 'dividend'}, default 'all'
            | Adjustment for pricing data.

            - all : apply both split and dividend adjustment
            - split : apply only split adjustment
            - dividend: apply only dividend adjustment
        currency : str, {'trade', 'report', ISO3 currency}, default 'trade'
            | Desired currency for the pricing data.

            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)
        package : str, {'Prism Market', 'CIQ Market', 'MI Integrated Market'}, default 'Prism Market'
            | Desired data package in where the pricing data outputs from.

            .. admonition:: Warning
                :class: warning

                If an invalid package is entered without a license, an error will be generated as output.

    Returns
    -------
        prism._PrismComponent: Historical time-series of daily end of day ask prices for stocks

    Examples
    --------
        >>> bid = prism.market.bid(adjustment='all', package='Prism Market')
        >>> bid_df = prism.get_data([bid], 'KOSPI 200 Index', startdate='2015-01-01', enddate='2020-12-31', shownid=['ticker'])
        >>> bid_df
                listingid        date            Bid   ticker
        0        20108718  2015-01-02    7616.379240  A004430
        1        20108718  2015-01-05    7864.843306  A004430
        2        20108718  2015-01-06    7883.955926  A004430
        3        20108718  2015-01-07    8304.433576  A004430
        4        20108718  2015-01-08    8161.088923  A004430
        ...           ...         ...            ...      ...
        298976  686744025  2020-12-23  151197.660000  A352820
        298977  686744025  2020-12-24  155644.650000  A352820
        298978  686744025  2020-12-28  154656.430000  A352820
        298979  686744025  2020-12-29  156632.870000  A352820
        298980  686744025  2020-12-30  157621.090000  A352820
    """
    @_validate_args
    def __init__(
        self,
        adjustment: _AdjustmentType = 'all',
        currency: _CurrencyTypeWithReportTrade = 'trade',
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))


class vwap(_PrismMarketComponent):
    """
    | Daily VWAP pricing history for equity securities.
    | Default frequency is business daily.

    Parameters
    ----------
        adjustment : str, {'all', 'split', 'dividend'}, default 'all'
            | Adjustment for pricing data.

            - all : apply both split and dividend adjustment
            - split : apply only split adjustment
            - dividend : apply only dividend adjustment
        currency : str, {'trade', 'report', ISO3 currency}, default 'trade'
            | Desired currency for the pricing data.

            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)
        package : str, {'CIQ Market'}, default 'CIQ Market'
            | Desired data package in where the pricing data outputs from.

            .. admonition:: Warning
                :class: warning

                If an invalid package is entered without a license, an error will be generated as output.

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> vwap = prism.market.ask(adjustment='all', package='CIQ Market')
        >>> vwap_df = prism.get_data([vwap], 'KOSPI 200 Index', startdate='2019-01-01', enddate='2020-12-31', shownid=['ticker'])
        >>> vwap_df
               listingid        date          VWAP   ticker
        0       20113302  2019-01-02  44884.613444  A078930
        1       20113302  2019-01-03  44146.922268  A078930
        2       20113302  2019-01-04  44883.697058  A078930
        3       20113302  2019-01-07  45929.294117  A078930
        4       20113302  2019-01-08  45773.508402  A078930
        ...          ...         ...           ...      ...
        97551  643903265  2020-12-23  14429.849563  A272210
        97552  643903265  2020-12-24  15179.657143  A272210
        97553  643903265  2020-12-28  16643.755102  A272210
        97554  643903265  2020-12-29  17252.000000  A272210
        97555  643903265  2020-12-30  17067.000000  A272210
    """
    @_validate_args
    def __init__(
        self,
        adjustment: _AdjustmentType = 'all',
        currency: _CurrencyTypeWithReportTrade = 'trade',
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))


class market_cap(_PrismMarketComponent):
    """
    | Market capitalization history for equity securities aggregated to the company level.
    | Default frequency is daily.

    Parameters
    ----------

        currency : str {'trade', 'report', ISO3 currency}, default 'trade'
            | Desired currency for the pricing data.

            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

        package : str {'CIQ Market'}, default 'CIQ Market'
            | Desired data package in where the pricing data outputs from.

            .. admonition:: Warning
                :class: warning

                If an invalid package is entered without a license, an error will be generated as output.


    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> mcap = prism.market.market_cap()
        >>> mcap_df = mcap.get_data(universe=1, startdate='2010-01-01', enddate='2015-12-31', shownid=['ticker'])
        >>> mcap_df
                listingid        date     marketcap   Ticker
        0        20108718  2010-06-11  2.592000e+11  A004430
        1        20108718  2010-06-12  2.592000e+11  A004430
        2        20108718  2010-06-13  2.592000e+11  A004430
        3        20108718  2010-06-14  2.592000e+11  A004430
        4        20108718  2010-06-15  2.640000e+11  A004430
        ...           ...         ...           ...      ...
        437100  278631846  2015-12-27  2.433650e+13  A028260
        437101  278631846  2015-12-28  2.317036e+13  A028260
        437102  278631846  2015-12-29  2.375490e+13  A028260
        437103  278631846  2015-12-30  2.341962e+13  A028260
        437104  278631846  2015-12-31  2.341962e+13  A028260
    """
    @_validate_args
    def __init__(
        self,
        currency: _CurrencyTypeWithReportTrade = 'trade',
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))


class volume(_PrismMarketComponent):
    """
    | Daily volume for equity securities.
    | Default frequency is business daily.

    Parameters
    ----------
        adjustment : str, {'all', 'split', 'dividend'}, default 'all'
            | Adjustment for pricing data.

            - all : apply both split and dividend adjustment
            - split : apply only split adjustment
            - dividend : apply only dividend adjustment

        package : str, {'Prism Market', 'Compustat', 'CIQ Market', 'MI Integrated Market'}, default 'Prism Market'
            | Desired data package in where the pricing data outputs from.

            .. admonition:: Warning
                :class: warning

                If an invalid package is entered without a license, an error will be generated as output.

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> volume = prism.market.volume(package='MI Integrated Market')
        >>> volume_df = prism.get_data([volume], 'KOSPI 200 Index', startdate='2019-01-01', enddate='2020-12-31', shownid=['ticker'])
        >>> volume_df
                listingid        date    volume   ticker
        0        20113302  2019-01-02  271000.0  A078930
        1        20113302  2019-01-03  178985.0  A078930
        2        20113302  2019-01-04  166248.0  A078930
        3        20113302  2019-01-07  118530.0  A078930
        4        20113302  2019-01-08  112217.0  A078930
        ...           ...         ...       ...      ...
        100638  686744025  2020-12-23  149780.0  A352820
        100639  686744025  2020-12-24  132606.0  A352820
        100640  686744025  2020-12-28  115532.0  A352820
        100641  686744025  2020-12-29  110664.0  A352820
        100642  686744025  2020-12-30  117489.0  A352820
    """
    @_validate_args
    def __init__(
        self,
        adjustment: _AdjustmentType = 'all',
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))


class dividend(_PrismMarketComponent):
    """
    | Dividend history for equity securities.
    | Default frequency is aperiodic daily.

    Parameters
    ----------
        adjustment : bool, default True
            | Whether to apply split adjustment for dividend data.

        currency : str, {'trade', 'report', ISO3 currency}, default 'trade'
            | Desired currency for the pricing data.

            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)
            - None : dividend payment currency

        aggregate: bool, default True
            | Desired aggregation for dividend. If True, dividends are aggregated based on listingid and exdate.

            - If `True`, paymentdate and dividendtype column will be dropped
            - If `True`, and currency is `None`, the currency will be automatically set to `trade`

        package : str, {'Prism Market', 'Compustat', 'CIQ Market'}, default 'Prism Market'
            | Desired data package in where the pricing data outputs from.

            .. admonition:: Warning
                :class: warning

                If an invalid package is entered without a license, an error will be generated as output.

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> dividend = prism.market.dividend(package='MI Integrated Market')
        >>> dividend _df = prism.get_data([dividend], 'KOSPI 200 Index', startdate='2019-01-01', enddate='2020-12-31', shownid=['ticker'])
        >>> dividend _df
              listingid        date  dividend  Ticker
        0       2586086  2010-02-11      0.14     AFL
        1       2586086  2010-05-17      0.14     AFL
        2       2586086  2010-08-16      0.14     AFL
        3       2586086  2010-11-15      0.15     AFL
        4       2586086  2011-02-11      0.15     AFL
                    ...         ...       ...     ...
        9544  344286611  2010-08-25      0.50     ITT
        9545  344286611  2010-11-09      0.50     ITT
        9546  344286611  2011-02-28      0.50     ITT
        9547  344286611  2011-05-18      0.50     ITT
        9548  344286611  2011-08-24      0.50     ITT
    """
    @_validate_args
    def __init__(
        self,
        adjustment: bool = True,
        currency: _CurrencyTypeWithReportTrade = 'trade',
        aggregate: bool = True,
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))


class dividend_adjustment_factor(_PrismMarketComponent):
    """
    | Dividend adjustment factor history for equity securities.
    | Default frequency is daily.

    Parameters
    ----------
    package : str, {'Prism Market', 'Compustat', 'CIQ Market'}, default 'Prism Market'
        | Desired data package in where the pricing data outputs from.

        .. admonition:: Warning
            :class: warning

            If an invalid package is entered without a license, an error will be generated as output.

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> divadj = prism.market.dividend_adjustment_factor()
        >>> divadj_df = divadj.get_data()
    """
    @_validate_args
    def __init__(self, package : str = None):
        super().__init__(**_get_params(vars()))


class split(_PrismMarketComponent):
    """
    | Return the split history for equity securities.
    | Default frequency is aperiodic daily.

    Parameters
    ----------
    package : str, {'Prism Market', 'Compustat', 'CIQ Market'}, default 'Prism Market'
        | Desired data package in where the pricing data outputs from.

        .. admonition:: Warning
            :class: warning

            If an invalid package is entered without a license, an error will be generated as output.

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> split = prism.market.split()
        >>> split_df = split.get_data(universe=1, startdate='2010-01-01', enddate='2015-12-31', shownid=['ticker'])
        >>> split_df
            listingid        date     split   Ticker
        0    20126254  2015-07-29  1.350000  A014830
        1    20158335  2013-04-30  0.100000  A008000
        2    20158445  2015-01-01  0.620935  A004150
        3    20158447  2010-12-29  1.030000  A001060
        4    20158758  2012-10-19  0.142857  A001440
        ..        ...         ...       ...      ...
        60  104646533  2014-09-01  0.478239  A060980
        61  107478344  2012-12-27  1.050000  A128940
        62  107478344  2014-02-10  1.050000  A128940
        63  107478344  2014-12-29  1.050000  A128940
        64  107478344  2015-12-29  1.020000  A128940
    """
    @_validate_args
    def __init__(self, package : str = None):
        super().__init__(**_get_params(vars()))


class split_adjustment_factor(_PrismMarketComponent):
    """
    | Split adjustment factor history for equity securities.
    | Default frequency is daily.

    Parameters
    ----------
    package : str, {'Prism Market', 'Compustat', 'CIQ Market'}, default 'Prism Market'
        | Desired data package in where the pricing data outputs from.

        .. admonition:: Warning
            :class: warning

            If an invalid package is entered without a license, an error will be generated as output.

    Returns
    -------
        prism._PrismComponent
    """
    @_validate_args
    def __init__(self, package : str = None):
        super().__init__(**_get_params(vars()))


class exchange_rate(_PrismMarketComponent):
    """
    | Daily exchange rate history.
    | Default frequency is daily.

    Parameters
    ----------
        currency : list of ISO3 currency
            | Desired exchange rates.
        to_convert : bool, default False
            | True
            | False : business daily
        package : str, {'Compustat', 'CIQ Market'}, default 'CIQ Market'
            | Desired data package in where the pricing data outputs from.

            .. admonition:: Warning
                :class: warning

                If an invalid package is entered without a license, an error will be generated as output.

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> exrt = prism.market.exchange_rate(currency=['USD', 'KRW'])
        >>> exrt_df = exrt.get_data(startdate='2010-01-01', enddate='2015-12-31')
        >>> exrt_df
            currency        date       exrt
        0         KRW  2010-01-01  1881.5000
        1         KRW  2010-01-02  1881.5000
        2         KRW  2010-01-03  1881.5000
        3         KRW  2010-01-04  1854.5000
        4         KRW  2010-01-05  1826.1000
        ...       ...         ...        ...
        3873      USD  2015-12-27     1.4933
        3874      USD  2015-12-28     1.4903
        3875      USD  2015-12-29     1.4795
        3876      USD  2015-12-30     1.4833
        3877      USD  2015-12-31     1.4740
    """
    @_validate_args
    def __init__(self, currency: list, to_convert: bool = False, package : str = None):
        super().__init__(**_get_params(vars()))

    @_validate_args
    @_req_call(_dataquery)
    def get_data(self, startdate: str = None, enddate: str = None, name = None,) -> pd.DataFrame: ...


class short_interest(_PrismMarketComponent):
    """
    | Short interest dataitems for equity securities and global data coverage.
    | Default frequency is business daily.

    Parameters
    ----------
    dataitemid : int
        | Unique identifier for the different dataitem. This identifies the type of the value (Revenue, Expense, etc.)

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> prism.market.short_interest.dataitems(search='short')
           dataitemid                               dataitemname  ...  datamodule                packagename
        0     1100035                Broker Short Interest Value  ...        None  IHS Markit Short Interest
        1     1100055        Short Interest Ratio (Day to Cover)  ...        None  IHS Markit Short Interest
        2     1100056                      Short Interest Tenure  ...        None  IHS Markit Short Interest
        3     1100057                       Short Interest Value  ...        None  IHS Markit Short Interest
        4     1100058          Short Interest as % Of Free Float  ...        None  IHS Markit Short Interest
        5     1100059  Short Interest as % Of Shares Outstanding  ...        None  IHS Markit Short Interest
        6     1100060                                Short Score  ...        None  IHS Markit Short Interest
        7     1100063           Supply Side Short Interest Value  ...        None  IHS Markit Short Interest

        >>> short = prism.market.short_interest(dataitemid=1100057)
        >>> short_df = short.get_data(universe=1, startdate='2010-01-01', enddate='2015-12-31', shownid=['ticker'])
        >>> short_df
                listingid        date  shortinterestvalue   Ticker
        0        20108718  2010-06-11        1.288440e+08  A004430
        1        20108718  2010-06-14        1.288440e+08  A004430
        2        20108718  2010-06-15        1.298000e+08  A004430
        3        20108718  2010-06-16        1.309800e+08  A004430
        4        20108718  2010-06-17        6.660000e+07  A004430
        ...           ...         ...                 ...      ...
        305527  278631846  2015-12-25        7.331920e+10  A028260
        305528  278631846  2015-12-28        7.045744e+10  A028260
        305529  278631846  2015-12-29        7.223796e+10  A028260
        305530  278631846  2015-12-30        6.464626e+10  A028260
        305531  278631846  2015-12-31        6.800626e+10  A028260
    """
    @_validate_args
    def __init__(self, dataitemid: int, currency: _CurrencyTypeWithReportTrade = 'trade', package : str = None):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search : str = None, package : str = None):
        """
        | Usable data items for the short_interest data component.

        Parameters
        ----------
            search : str, default None
                | Search word for data items name, the search is case-insensitive.
            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to short_interest data component.

            Columns:
                - *dataitemid : int*
                - *dataitemname : str*
                - *dataitemdescription : str*
                - *datamodule : str*
                - *datacomponent : str*
                - *packagename : str*

        Examples
        --------
            >>> prism.market.short_interest.dataitems(search='short')
            dataitemid                                  dataitemname  ...  datamodule                packagename
            0     1100035                Broker Short Interest Value  ...        None  IHS Markit Short Interest
            1     1100055        Short Interest Ratio (Day to Cover)  ...        None  IHS Markit Short Interest
            2     1100056                      Short Interest Tenure  ...        None  IHS Markit Short Interest
            3     1100057                       Short Interest Value  ...        None  IHS Markit Short Interest
            4     1100058          Short Interest as % Of Free Float  ...        None  IHS Markit Short Interest
            5     1100059  Short Interest as % Of Shares Outstanding  ...        None  IHS Markit Short Interest
            6     1100060                                Short Score  ...        None  IHS Markit Short Interest
            7     1100063           Supply Side Short Interest Value  ...        None  IHS Markit Short Interest
        """
        return _list_dataitem(
            datacategoryid=cls.categoryid,
            datacomponentid=cls.componentid,
            search=search,
            package=package,
        )


class shares_outstanding(_PrismMarketComponent):
    """
    | The total number of shares of a security that are currently held by all its shareholders and are available for trading on a specific stock exchange.
    | Default frequency is daily.

    Parameters
    ----------
    adjustment : bool, default True
        | Adjustment for Shares outstanding

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> shares_out = prism.market.shares_outstanding()
        >>> shares_out.get_data("Semiconductor", "2020-01-01", shownid=['Company Name'])
                  listingid        date      shares                                       Company Name
        0           2586491  2020-01-01    39521782                                          AXT, Inc.
        1           2587243  2020-01-01    22740986                                  Aehr Test Systems
        2           2587303  2020-01-01  1138599272                       Advanced Micro Devices, Inc.
        3           2587347  2020-01-01    38308569                   Advanced Energy Industries, Inc.
        4           2589783  2020-01-01   239783075                             Amkor Technology, Inc.
        ...             ...         ...         ...                                                ...
        1168126  1831385562  2023-07-26     7501500                                        SEALSQ Corp
        1168127  1833187092  2023-07-26    12675758                                  GigaVis Co., Ltd.
        1168128  1833609849  2023-07-26  7021800000  Semiconductor Manufacturing Electronics (Shaox...
        1168129  1834641950  2023-07-26   452506348     Smarter Microelectronics (Guangzhou) Co., Ltd.
        1168130  1838168164  2023-07-26    37844925              Integrated Solutions Technology, Inc.
    """
    @_validate_args
    def __init__(self, adjustment: bool = True, package : str = None):
        super().__init__(**_get_params(vars()))


class enterprise_value(_PrismMarketComponent):
    """
    | Represents the total value of a company, taking into account its market capitalization, outstanding debt, cash, and other financial assets.
    | It is used to determine the true cost of acquiring a company and is calculated as market capitalization plus total debt minus cash and cash equivalents.
    | Default frequency is daily.

    Parameters
    ----------
    currency : str, {'trade', 'report', ISO3 currency}, default None
            | Desired currency for the enterprise value data.

            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)
            - None : dividend payment currency

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> ent_val = prism.market.enterprise_value(currency='trade')
        >>> ent_val.get_data("Semi", "2020-01-01", shownid=['Company Name'])
                  listingid        date           tev  currency                                       Company Name
        0           2586491  2020-01-01  1.546798e+08       USD                                          AXT, Inc.
        1           2587243  2020-01-01  4.288797e+07       USD                                  Aehr Test Systems
        2           2587303  2020-01-01  5.211816e+10       USD                       Advanced Micro Devices, Inc.
        3           2587347  2020-01-01  2.842980e+09       USD                   Advanced Energy Industries, Inc.
        4           2589783  2020-01-01  3.988825e+09       USD                             Amkor Technology, Inc.
        ...             ...         ...           ...       ...                                                ...
        1167719  1831385562  2023-07-25  1.645610e+08       USD                                        SEALSQ Corp
        1167720  1833187092  2023-07-25           NaN       KRW                                  GigaVis Co., Ltd.
        1167721  1833609849  2023-07-25  5.481000e+10       CNY  Semiconductor Manufacturing Electronics (Shaox...
        1167722  1834641950  2023-07-25  9.114210e+09       CNY     Smarter Microelectronics (Guangzhou) Co., Ltd.
        1167723  1838168164  2023-07-25           NaN       TWD              Integrated Solutions Technology, Inc.
    """
    @_validate_args
    def __init__(self, currency: _CurrencyTypeWithReportTrade = None, package : str = None):
        super().__init__(**_get_params(vars()))


class implied_market_cap(_PrismMarketComponent):
    """
    | Theoretical total value of a company's outstanding shares, including potential dilution from stock options and other equity-based compensation plans.
    | Default frequency is daily.

    Parameters
    ----------
    dilution : str, {'all', 'partner', 'exercisable'}, default 'all'
            | Options whether to include which potential dilution from stock options and other equity-based compensation plans.

    currency : str, {'trade', 'report', ISO3 currency}, default None
            | Desired currency for the pricing data.

            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)
            - None : dividend payment currency

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> immcap = prism.market.implied_market_cap()
        >>> immcap.get_data("Semi", "2020-01-01", shownid=['Company Name'])
             listingid        date    implieddilutedmarketcapout  currency               Company Name
        0    707934944  2021-06-24                  1.317918e+09       USD  indie Semiconductor, Inc.
        1    707934944  2021-06-25                  1.335508e+09       USD  indie Semiconductor, Inc.
        2    707934944  2021-06-26                  1.335508e+09       USD  indie Semiconductor, Inc.
        3    707934944  2021-06-27                  1.335508e+09       USD  indie Semiconductor, Inc.
        4    707934944  2021-06-28                  1.332802e+09       USD  indie Semiconductor, Inc.
        ...        ...         ...                           ...       ...                        ...
        762  707934944  2023-07-26                  1.485848e+09       USD  indie Semiconductor, Inc.
        763  707934944  2023-07-27                  1.464184e+09       USD  indie Semiconductor, Inc.
        764  707934944  2023-07-28                  1.502095e+09       USD  indie Semiconductor, Inc.
        765  707934944  2023-07-29                  1.502095e+09       USD  indie Semiconductor, Inc.
        766  707934944  2023-07-30                  1.502095e+09       USD  indie Semiconductor, Inc.


    """
    @_validate_args
    def __init__(self, dilution: _DilutionType = 'all', currency: _CurrencyTypeWithReportTrade = None, package : str = None):
        super().__init__(**_get_params(vars()))


@_validate_args
def dataitems(search : str = None, package : str = None):
    """
    | Usable dataitems for the market data categories.

    Parameters
    ----------
        search : str, default None
            | Search word for Data Items name, the search is case-insensitive.
        package : str, default None
            | Search word for package name, the search is case-insensitive.

    Returns
    -------
        pandas.DataFrame
            Data items that belong to market data categories.

        Columns:
            - *dataitemid : int*
            - *dataitemname : str*
            - *dataitemdescription : str*
            - *datamodule : str*
            - *datacomponent : str*
            - *packagename : str*

    Examples
    --------
        >>> prism.market.dataitems(search='short')
            dataitemid                               dataitemname  ...   datacomponent                packagename
        0     1100035                Broker Short Interest Value  ...  Short Interest  IHS Markit Short Interest
        1     1100055        Short Interest Ratio (Day to Cover)  ...  Short Interest  IHS Markit Short Interest
        2     1100056                      Short Interest Tenure  ...  Short Interest  IHS Markit Short Interest
        3     1100057                       Short Interest Value  ...  Short Interest  IHS Markit Short Interest
        4     1100058          Short Interest as % Of Free Float  ...  Short Interest  IHS Markit Short Interest
        5     1100059  Short Interest as % Of Shares Outstanding  ...  Short Interest  IHS Markit Short Interest
        6     1100060                                Short Score  ...  Short Interest  IHS Markit Short Interest
        7     1100063           Supply Side Short Interest Value  ...  Short Interest  IHS Markit Short Interest
    """
    return _list_dataitem(
        datacategory=_PrismMarketComponent.categoryid,
        datacomponent=None,
        search=search,
        package=package,
    )



from ..._prismcomponent.prismcomponent import _PrismComponent, _PrismDataComponent
from .._req_builder import _list_dataitem
from ..._utils import _validate_args

__all__ = ['alpha_factor_library', 'dataitems']


_data_category = __name__.split(".")[-1]


@_validate_args
class alpha_factor_library(_PrismComponent, _PrismDataComponent):
    """
    | Pre-calculated factor dataset provided by S&P Global Market Intelligence.
    | Default frequency is quarterly.


    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the value (Revenue, Expense, etc.)

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> di = prism.precalculated.afl.dataitems('Turnover')
        >>> di[['dataitemid', 'dataitemname']]
           dataitemid                                  dataitemname
        0      900075                      1Y Chg in Sales Turnover
        1      900078                      1Y Chg in Share Turnover
        2      900114                      3Y Chg in Sales Turnover
        3      900158      5 Yr Hist Rel Receivables Turnover Ratio
        4	   900168  5 Yr Hist Rel Working Capital Turnover Ratio
        5      900225                                Asset Turnover
        6      900226                             Asset Turnover v2
        7      900261                      Change of Asset Turnover
        8      900262                   Change of Asset Turnover v2
        9      900306                               Equity Turnover
        10     900352                    Ind Grp Rel Asset Turnover
        11     900384                Ind Grp Rel Inventory Turnover
        12     900404        Ind Grp Rel Receivables Turnover Ratio
        13     900420    Ind Grp Rel Working Capital Turnover Ratio
        14     900431                Institution Ownership Turnover
        15     900434                            Inventory Turnover
        16     900525                    Receivables Turnover Ratio
        17     900568                                Share Turnover
        18     900636                Working Capital Turnover Ratio

        >>> turnover = prism.precalculated.alpha_factor_library(dataitemid=900306)
        >>> turnover_df = turnover.get_data(universe='S&P 500', startdate='2010-01-01', enddate='2015-12-31', shownid=['ticker'])
        >>> turnover_df
              listingid        date  Equity Turnover  Ticker
        0       2586086  2010-01-31          2.74645     AFL
        1       2586086  2010-02-28          2.61951     AFL
        2       2586086  2010-03-31          2.61951     AFL
        3       2586086  2010-04-30          2.33898     AFL
        4       2586086  2010-05-31          2.33898     AFL
          ...         ...         ...              ...     ...
        36623  344286611  2011-06-30          2.53677    ITT
        36624  344286611  2011-07-31          2.46787    ITT
        36625  344286611  2011-08-31          2.46787    ITT
        36626  344286611  2011-09-30          2.46787    ITT
        36627  344286611  2011-10-31          2.55218    ITT
    """

    _component_category_repr = _data_category

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None):
        """
        Usable data items for the alpha factor library datacomponent.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:
                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> di = prism.precalculated.afl.dataitems("Debt")
            >>> di[['dataitemid', 'dataitemname']]
            dataitemid                                       dataitemname
            0      900033  1-Year Change in Long Term Debt to Avg Total A...
            1      900059                           1Y Chg in Debt to Assets
            2      900145       5 Yr Hist Rel Long Term Debt to Assets Ratio
            3      900279                               Debt to Assets Ratio
            4      900366                   Ind Grp Rel Debt to Assets Ratio
            5      900388         Ind Grp Rel Long Term Debt to Assets Ratio
            6      900389         Ind Grp Rel Long Term Debt to Equity Ratio
            7      900425    Ind Grp Rel Year over Year Change of Total Debt
            8      900451                     Long Term Debt to Assets Ratio
            9      900452                     Long Term Debt to Equity Ratio
            10     900644                Year over Year Change of Total Debt
        """
        return _list_dataitem(
            datacategoryid=cls.categoryid,
            datacomponentid=cls.componentid,
            search=search,
            package=None,
        )


def dataitems(search: str = None):
    """
    Usable data items for the precalculated data category.

    Parameters
    ----------
        search : str, default None
            | Search word for dataitems name, the search is case-insensitive.

    Returns
    -------
        pandas.DataFrame
            Data items that belong to cash flow statement data component.

        Columns:
            - *datamodule*
            - *datacomponent*
            - *dataitemid*
            - *datadescription*

    Examples
    --------
        >>> di = prism.precalculated.dataitems("Debt")
        >>> di[['dataitemid', 'dataitemname']]
           dataitemid                                       dataitemname
        0      900033  1-Year Change in Long Term Debt to Avg Total A...
        1      900059                           1Y Chg in Debt to Assets
        2      900145       5 Yr Hist Rel Long Term Debt to Assets Ratio
        3      900279                               Debt to Assets Ratio
        4      900366                   Ind Grp Rel Debt to Assets Ratio
        5      900388         Ind Grp Rel Long Term Debt to Assets Ratio
        6      900389         Ind Grp Rel Long Term Debt to Equity Ratio
        7      900425    Ind Grp Rel Year over Year Change of Total Debt
        8      900451                     Long Term Debt to Assets Ratio
        9      900452                     Long Term Debt to Equity Ratio
        10     900644                Year over Year Change of Total Debt
    """
    return _list_dataitem(
        datacategoryid=alpha_factor_library.categoryid,
        datacomponentid=None,
        search=search,
        package=None,
    )



from prism._utils.exceptions import PrismValueError
from ..._common.config import *
from ..._common.const import *
from ..._utils import _validate_args, post
from ..._utils.validate_utils import get_sm_attributevalue


__all__ = ['get_securitymaster', 'get_securitymaster_advanced']


@_validate_args
def get_securitymaster(attribute: str, search: str):
    """
    Return Security information search result with single condition.

    Parameters
    ----------
        attribute : str
            Security Master attribute to search on.

        search : str
            Search word to condition on the attribute.

    Returns
    -------
        pandas.Dataframe
            | Security information.
            | Columns:

            - *listingid*
            - *valuetype*
            - *value*
            - *startdate*
            - *enddate*

    Examples
    --------
        >>> prism.get_securitymaster(attribute='companyname', search='Samsung')
            listingid      valuetype     value	 startdate     enddate
        0       2647420  tradingitemid   2647420  1700-01-01  2199-12-31
        1      20174680  tradingitemid  20174680  1700-01-01  2199-12-31
        2      30562725  tradingitemid  30562725  1700-01-01  2199-12-31
        3       2647422  tradingitemid   2647422  1700-01-01  2199-12-31
        4      30562723  tradingitemid  30562723  1700-01-01  2199-12-31
        ...	        ...            ...       ...         ...         ...
        5559   20194961  fsym_entityid  05HXF2-E  1700-01-01  2199-12-31
        5560   20191195  fsym_entityid  05HWPG-E  1700-01-01  2199-12-31
        5561   31778919  fsym_entityid  05VHCF-E  1700-01-01  2199-12-31
        5562   62339029  fsym_entityid  07LZ7F-E  1700-01-01  2199-12-31
        5563   62339030  fsym_entityid  07LZ7F-E  1700-01-01  2199-12-31
    """
    attributevalue = get_sm_attributevalue(attribute)
    query = [{'attribute': attributevalue, 'search': search}]
    return post(URL_SM, None, query)


@_validate_args
def get_securitymaster_advanced(queries: list):
    """
    Return Security information search result with multiple conditions.

    Parameters
    ----------
        queries: list of dict
            | List of security information search criteria. Each dictionary contains three keys: “operator”, “attribute”, “search”

            - operator: {”AND”, “OR”}
            - attribute: Security Master attribute to search on.
            - search: Search word to condition on the attribute.

    Returns
    -------
        pandas.Dataframe
            | Security information
            | Columns:

            - *listingid*
            - *valuetype*
            - *value*
            - *startdate*
            - *enddate*

    Examples
    --------
        >>> prism.get_securitymaster_advanced([{'attribute': 'country', 'search': 'US'}, {'attribute': 'GICS sector', 'search': '30'}])
                listingid      valuetype     value   startdate     enddate
        0         2585879  tradingitemid   2585879  1987-05-08  1993-10-15
        1         2587041  tradingitemid   2587041  1983-09-20  2002-09-13
        2         2588957  tradingitemid   2588957  1984-10-22  1993-02-25
        3         2588959  tradingitemid   2588959  1984-10-22  1993-02-25
        4         2587848  tradingitemid   2587848  1986-04-23  2003-11-05
        ...        ...            ...       ...         ...         ...
        115323   27580549  fsym_entityid  0044NF-E  1700-01-01  2199-12-31
        115324   27580549  fsym_entityid  06GRML-E  1700-01-01  2199-12-31
        115325   40843715  fsym_entityid  000LNJ-E  1700-01-01  2199-12-31
        115326   99731436  fsym_entityid  05N251-E  1700-01-01  2199-12-31
        115327  144070140  fsym_entityid  00D843-E  1700-01-01  2199-12-31
    """
    if len(queries) == 0: raise PrismValueError('At least one query must be provided.')
    for i in range(len(queries)):
        assert isinstance(queries[i], dict), 'queries must be list of dicts.'
        assert set(queries[i].keys()) - {'operator'} == {'attribute', 'search'}, \
            'Valid arguments are "attribute", "search", and "operator".'
        queries[i]['attribute'] = get_sm_attributevalue(queries[i]['attribute'])

    return post(URL_SM, None, queries)



import pytest

from prism._common.const import FinancialPreliminaryType, FinancialPeriodType as _PeriodType
from prism._utils.validate_utils import _validate_args
from prism._utils.exceptions import PrismTypeError, PrismValueError


@_validate_args
def get_balance_sheet(
        preliminary: FinancialPreliminaryType,
        universe=None,
        dataitemid: int = 1,
        periodtype: _PeriodType = 'A',
        engine=None,
        offset: int = 365,
        n_quarter: int = 1,
        currency: str = 'trade') -> FinancialPreliminaryType:
    """
    valid argument:
        'keep', 'ignore', 'drop', 'null'
    """
    return preliminary


def test_preliminary_param():
    assert get_balance_sheet(preliminary=None) is None
    assert get_balance_sheet(universe='S&P 500', dataitemid=101, periodtype='A',
                             engine=None, offset=180, n_quarter=1, preliminary='keep', currency='trade') == FinancialPreliminaryType.KEEP

    # check type
    # assert isinstance(get_balance_sheet(preliminary='keep'), FinancialPreliminaryType)

    # valid preliminary param
    assert get_balance_sheet(
        preliminary='keep') == FinancialPreliminaryType.KEEP
    assert get_balance_sheet(
        preliminary='ignore') == FinancialPreliminaryType.IGNORE
    assert get_balance_sheet(
        preliminary='drop') == FinancialPreliminaryType.DROP
    assert get_balance_sheet(
        preliminary='null') == FinancialPreliminaryType.NULL


def test_wrong_preliminary_param():
    # wrong value
    with pytest.raises(Exception):
        assert get_balance_sheet(
            preliminary="KEEP") == FinancialPreliminaryType.KEEP

    with pytest.raises(Exception):
        assert get_balance_sheet(preliminary="do sth")

    # wrong type
    with pytest.raises(Exception):
        assert get_balance_sheet(preliminary=1)



from ..._prismcomponent.prismcomponent import _PrismModelComponent, _AbstractPrismComponent
from ..._utils import _validate_args


_data_category = __name__.split(".")[-1]


class _PrismTCModelComponent(_PrismModelComponent):
    _component_category_repr  = _data_category


class almgren(_PrismTCModelComponent):
    r"""
    | Estimates and manage the trading costs associated with executing large orders in financial markets for a given trade.
    | The model was created/calibrated based on the data from Citi's trading desk. It uses 3/5 power laws instead of conventional square-root law.
    |
    .. math:: C = \frac 1 2 + \eta\sigma|\frac X {VT}|^\frac 3 5
    .. math:: I = \sigma\gamma \frac X V(\frac \theta V)^\frac 1 4
    |
    .. math:: X \equiv \text {Transactions in shares}
    .. math:: T \equiv \text {Trade Duration}
    .. math:: V \equiv \text {Average Daily Volume}
    .. math:: \theta \equiv \text {Shares Outstanding}
    .. math:: \gamma = 0.314
    .. math:: \eta = 0.142


    Parameters
    ----------

        transaction : prism._PrismComponent
            | Desired currency for the pricing data.

        tradeduration : int
            | Desired data package in where the pricing data outputs from.

        averageperiod : int
            | Desired data package in where the pricing data outputs from.

    Returns
    -------
        prism._PrismComponent

    References
    ----------
        Almgren, Robert & Thum, Chee & Hauptmann, Emmanuel & Li, Hong. (2005). Direct Estimation of Equity Market Impact. RISK. 18.

    Examples
    --------
        >>> ret = prism.market.close().n_periods_pct_change(1)
        >>> trade = (ret > 0).map({True:1000_000, False:-1000_000})
        >>> trade_amount = trade.abs()
        >>> prism.tcmodel.almgren(trade_amount, 10, 5).get_data("Korea_primary", "2020-01-01", shownid=["companyname"])
                  listingid        date     value                         Company Name
        0          20108704  2020-01-02  0.235841      Daewon Pharmaceutical Co., Ltd.
        1          20108706  2020-01-02  0.344619                  N.I Steel Co., Ltd.
        2          20108718  2020-01-02  0.046685         Songwon Industrial Co., Ltd.
        3          20108719  2020-01-02  0.275940               Whan In Pharm Co.,Ltd.
        4          20109325  2020-01-02  0.030277               Theragen Etex Co.,Ltd.
        ...             ...         ...       ...                                  ...
        1611478  1833210191  2023-07-28  0.038329                    Suresofttech Inc.
        1611479  1833230315  2023-07-28  0.111412                S.Biomedics CO., LTD.
        1611480  1834641955  2023-07-28  0.034584     ISU Specialty Chemical Co., Ltd.
        1611481  1838438397  2023-07-28  0.069874   Dongkuk Steel Mill Company Limited
        1611482  1838459712  2023-07-28  0.127869                 Dongkuk CM Co., Ltd.
    """
    @_validate_args
    def __init__(
        self,
        transaction: _AbstractPrismComponent,
        tradeduration: int,
        averageperiod: int,
    ):
        super().__init__(
            transaction=transaction,
            tradeduration=tradeduration,
            averageperiod=averageperiod,
            children=[transaction],
        )


class bidaskspread(_PrismTCModelComponent):
    r"""
    | The simple bid-ask spread market impact model calculates the relative impact of a trade.
    | The model provides a basic measure of liquidity and reflects the cost a trader may incur when buying or selling an asset, with a larger spread indicating a higher potential market impact.
    |
    .. math:: \frac {ask-bid} {(ask+bid)/2}

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> bas = prism.tcmodel.bidaskspread()
        >>> bas.get_data("Korea_primary", "2020-01-01", shownid=["companyname"])
                  listingid        date  bidaskspread                        Company Name
        0          20108704  2020-01-02      0.002894     Daewon Pharmaceutical Co., Ltd.
        1          20108706  2020-01-02      0.009009                 N.I Steel Co., Ltd.
        2          20108718  2020-01-02      0.003221        Songwon Industrial Co., Ltd.
        3          20108719  2020-01-02      0.003190              Whan In Pharm Co.,Ltd.
        4          20109325  2020-01-02      0.002410              Theragen Etex Co.,Ltd.
        ...             ...         ...           ...                                 ...
        1611958  1833210191  2023-07-28      0.002315                   Suresofttech Inc.
        1611959  1833230315  2023-07-28      0.003415               S.Biomedics CO., LTD.
        1611960  1834641955  2023-07-28      0.001857    ISU Specialty Chemical Co., Ltd.
        1611961  1838438397  2023-07-28      0.001007  Dongkuk Steel Mill Company Limited
        1611962  1838459712  2023-07-28      0.001129                Dongkuk CM Co., Ltd.
    """



import numpy as np
import pandas as pd
from enum import Enum


__all__ = [
    # true const
    'BEGINNING_DATE',
    'ACTIVE_DATE',
    'PrismComponentType',
    'SMValues',
    'CategoryComponent',
    'PACKAGE_NAME',
    'SPECIALVALUEMAP',
    'FILEEXTENSION',
    # param types
    'PreferenceType',
    'FrequencyType',
    'ScreenFrequencyType',
    'FBFrequencyType',
    'UniverseFrequencyType',
    'FinancialPeriodType',
    'EstimatePeriodType',
    'AdjustmentType',
    'DilutionType',
    'RankType',
    'DateType',
    'AggregationType',
    'FinancialPreliminaryType',
    'FillnaMethodType',
    'CompanyRelAttributeType',
]


BEGINNING_DATE = pd.to_datetime('1700-01-01')
ACTIVE_DATE = pd.to_datetime('2199-12-31')
SMValues = None
PreferenceType = None
CategoryComponent = None
FunctionComponents = None
TaskComponents = None
DataComponents = None


class PrismComponentType(str, Enum):
    FUNCTION_COMPONENT = 'functioncomponent'
    DATA_COMPONENT = 'datacomponent'
    TASK_COMPONENT = 'taskcomponent'
    MODEL_COMPONENT = 'modelcomponent'


class AdjustmentType(Enum):
    ALL = 'all'
    SPLIT = 'split'
    DIVIDEND = 'dividend'
    NONE = None

class CompanyRelAttributeType(Enum):
    COMPANYNAME = "companyname"
    LISTINGID = "listingid"
    COMPANYID = "companyid"


class DilutionType(Enum):
    ALL = "all"
    PARTNER = "partner"
    EXERCISABLE = "exercisable"


class FrequencyType(str, Enum):
    NANOSECONDS = 'N'
    MICROSECONDS = 'U'
    MICROSECONDS_ALIAS = 'us'
    MILISECONDS = 'L'
    MILISECONDS_ALIAS = 'ms'
    SECONDS = 'S'
    MINUTES = 'T'
    MINUTES_ALIAS = 'min'
    HOURS = 'H'
    BUSINESS_HOURS = 'BH'
    CALENDAR_DAY = 'D'
    BUSINESS_DAY = 'BD'
    WEEKS = 'W'
    MONTH_START = 'MS'
    BUSINESS_MONTH_START = 'BMS'
    SEMI_MONTH_START = 'SMS'
    SEMI_MONTH_END = 'SM'
    BUSINESS_MONTH_END = 'BM'
    MONTH_END = 'M'
    QUARTER_END = 'Q'
    QUARTER_START = 'QS'
    BUSINESS_QUARTER_END = 'BQ'
    BUSINESS_QUARTER_START = 'BQS'
    YEAR_START = 'AS'
    YEAR_END = 'A'


class ResampleFrequencyType(str, Enum):
    CALENDAR_DAY = 'D'
    BUSINESS_DAY = 'BD'
    WEEKS = 'W'
    BUSINESS_MONTH_END = 'BM'
    MONTH_END = 'M'
    QUARTER_END = 'Q'
    YEAR_END = 'A'


class ScreenFrequencyType(str, Enum):
    CALENDAR_DAY = 'D'
    BUSINESS_DAY = 'BD'
    WEEKS = 'W'
    BUSINESS_MONTH_END = 'BM'
    MONTH_END = 'M'
    QUARTER_END = 'Q'
    YEAR_END = 'A'


class FBFrequencyType(str, Enum):
    CALENDAR_DAY = 'D'
    BUSINESS_DAY = 'BD'
    WEEKS = 'W'
    BUSINESS_MONTH_END = 'BM'
    MONTH_END = 'M'
    QUARTER_END = 'Q'
    YEAR_END = 'A'


class UniverseFrequencyType(str, Enum):
    CALENDAR_DAY = 'D'
    WEEKS = 'W'
    MONTH_START = 'MS'
    SEMI_MONTH_START = 'SMS'
    SEMI_MONTH_END = 'SM'
    MONTH_END = 'M'
    QUARTER_END = 'Q'
    QUARTER_START = 'QS'
    YEAR_START = 'AS'
    YEAR_END = 'A'


class RankType(str, Enum):
    STANDARD = 'standard'
    MODIFIED = 'modified'
    DENSE = 'dense'
    ORDINAL = 'ordinal'
    FRACTIONAL = 'fractional'


class DateType(str, Enum):
    ENTEREDDATE = 'entereddate'
    ANNOUNCEDDATE = 'announceddate'


class FinancialPeriodType(str, Enum):
    ANNUAL = 'Annual'
    A = 'A'
    SEMI_ANNUAL = 'Semi-Annual'
    SA = 'SA'
    QUARTERLY = 'Quarterly'
    Q = 'Q'
    YTD = 'YTD'
    LTM = 'LTM'
    QSA = 'Q-SA'


class EstimatePeriodType(str, Enum):
    ANNUAL = 'Annual'
    A = 'A'
    SEMI_ANNUAL = 'Semi-Annual'
    SA = 'SA'
    QUARTERLY = 'Quarterly'
    Q = 'Q'
    NTM = 'NTM'
    QSA = 'Q-SA'
    NONE = None


class AggregationType(str, Enum):
    ONEDAY = '1 day'
    ONEWEEK = '1 week'
    ONEMONTH = '1 month'
    TWOMONTH = '2 month'
    THREEMONTH = '3 month'
    THREEMONTHLATEST = '3 month latest'


class FinancialPreliminaryType(str, Enum):
    KEEP = 'keep'
    IGNORE = 'ignore'
    NULL = 'null'


class FillnaMethodType(str, Enum):
    BACKFILL = 'backfill'
    BFILL = 'bfill'
    PAD = 'pad'
    FFILL = 'ffill'
    NONE = None


FILEEXTENSION = {'pdq': 'dataquery', 'ptq': 'taskquery', 'pws': 'workspace', 'puv': 'universe', 'ppt': 'portfolio', 'ped': 'datafile'}


# PACKAGE_NAME = 'p3s9'
PACKAGE_NAME = 'prism'


SPECIALVALUEMAP = {
    np.nan: "\x01NaN",
    np.inf: "\x01inf",
    np.NINF: "\x01ninf",
}


DaysRepr = [
    'm',
    'mon',
    'monday',
    't',
    'tu',
    'tue',
    'tues',
    'tuesday',
    'w',
    'wed',
    'wednesday',
    'r',
    'th',
    'thu',
    'thurs',
    'thursday',
    'f',
    'fr',
    'fri',
    'friday',
    's',
    'sa',
    'sat',
    'saturday',
    'u',
    'su',
    'sun',
    'sunday'
]

CURRENCYLIST = [
    "AAD",
    "ADP",
    "AED",
    "AFA",
    "AFN",
    "ALL",
    "AMD",
    "ANG",
    "AOA",
    "AON",
    "ARS",
    "ATS",
    "AUD",
    "AWG",
    "AZM",
    "AZN",
    "BAM",
    "BBD",
    "BDT",
    "BEF",
    "BGL",
    "BGN",
    "BHD",
    "BIF",
    "BMD",
    "BND",
    "BOB",
    "BRL",
    "BSD",
    "BTN",
    "BWP",
    "BYN",
    "BYR",
    "BZD",
    "CAD",
    "CDF",
    "CHF",
    "CLF",
    "CLP",
    "CNH",
    "CNY",
    "COP",
    "CRC",
    "CSD",
    "CUP",
    "CVE",
    "CYP",
    "CZK",
    "DDM",
    "DEM",
    "DJF",
    "DKK",
    "DOP",
    "DZD",
    "ECS",
    "EEK",
    "EGP",
    "ERN",
    "ESP",
    "ETB",
    "EUR",
    "FIM",
    "FJD",
    "FKP",
    "FRF",
    "GBP",
    "GBX",
    "GEL",
    "GHC",
    "GHS",
    "GIP",
    "GMD",
    "GNF",
    "GRD",
    "GTQ",
    "GWP",
    "GYD",
    "HKD",
    "HNL",
    "HRK",
    "HTG",
    "HUF",
    "IDN",
    "IDR",
    "IEP",
    "ILA",
    "ILS",
    "INR",
    "IQD",
    "IRR",
    "ISK",
    "ITL",
    "JMD",
    "JOD",
    "JPY",
    "KES",
    "KGS",
    "KHR",
    "KMF",
    "KPW",
    "KRW",
    "KWD",
    "KWF",
    "KYD",
    "KZT",
    "LAK",
    "LBP",
    "LKR",
    "LRD",
    "LSL",
    "LTL",
    "LUF",
    "LVL",
    "LYD",
    "MAD",
    "MDL",
    "MGA",
    "MGF",
    "MKD",
    "MMK",
    "MNT",
    "MOP",
    "MRO",
    "MRU",
    "MTL",
    "MUR",
    "MVR",
    "MWK",
    "MXN",
    "MXV",
    "MYR",
    "MZM",
    "MZN",
    "NAD",
    "NGN",
    "NIO",
    "NLG",
    "NOK",
    "NPR",
    "NZD",
    "OMR",
    "PAB",
    "PEN",
    "PGK",
    "PHP",
    "PKR",
    "PLN",
    "PPP",
    "PTE",
    "PYG",
    "QAR",
    "ROL",
    "RON",
    "RSD",
    "RUB",
    "RUR",
    "RWF",
    "SAC",
    "SAR",
    "SBD",
    "SCR",
    "SDD",
    "SDG",
    "SDP",
    "SEK",
    "SGD",
    "SHP",
    "SIT",
    "SKK",
    "SLL",
    "SOS",
    "SRD",
    "SRG",
    "STD",
    "SVC",
    "SYP",
    "SZL",
    "THB",
    "TJR",
    "TJS",
    "TMM",
    "TMT",
    "TND",
    "TOF",
    "TOP",
    "TPE",
    "TRL",
    "TRY",
    "TTD",
    "TWD",
    "TZS",
    "UAH",
    "UGX",
    "USD",
    "USX",
    "UYU",
    "UZS",
    "VEB",
    "VEF",
    "VES",
    "VND",
    "VUV",
    "WST",
    "XAF",
    "XBA",
    "XBB",
    "XCD",
    "XDR",
    "XEU",
    "XOF",
    "XPF",
    "XUA",
    "YER",
    "YUM",
    "ZAC",
    "ZAR",
    "ZMK",
    "ZMW",
    "ZWD",
    "ZWL",
]

CurrencyType = Enum('CurrencyType', [(c, c) for c in CURRENCYLIST], type=str)

CurrencyTypeWithReportTrade = Enum('CurrencyTypeWithReportTrade', [(c, c) for c in ["report", "trade",] + CURRENCYLIST], type=str)




from .prismcomponent import _PrismComponent, _PrismFinancialComponent
from .abstract_prismcomponent import _AbstractPrismComponent

__all__ = [
    "_AbstractPrismComponent",
    "_PrismComponent",
    "_PrismFinancialComponent",
]


# Use this code snippet in your app.
# If you need more information about configurations
# or implementing the sample code, visit the AWS docs:
# https://aws.amazon.com/developer/language/python/

import re
import boto3
import orjson
from sqlalchemy import select
from botocore.exceptions import ClientError

from tests.test_util.dq_model import PrismQuery
from tests.test_util.dq_table import dataqueries


def query_test(queries, engine):
    for k, v in queries.items():
        dq = engine.execute(select(dataqueries.c.dataquerybody).where(dataqueries.c.dataqueryname == k)).fetchone()
        cross_check(v._query, dq[0])


def get_secret(secret_name):
    session = boto3.session.Session()
    client = session.client(
        service_name='secretsmanager',
        region_name="ap-northeast-2"
    )

    try:
        get_secret_value_response = client.get_secret_value(
            SecretId=secret_name
        )
    except ClientError as e:
        raise e

    secret = get_secret_value_response['SecretString']
    return orjson.loads(secret)


def replace_nodeid(json_string):
    regex = ',[ ]?"nodeid":[ ]?"[^,]*"'
    replace = ''
    tmp = re.sub(regex, replace, json_string)
    regex = '"nodeid":[ ]?"[^\}\],]*",'
    tmp = re.sub(regex, replace, tmp)
    regex = ', "nodeid":[ ]?"[^\}\],]*"'
    tmp = re.sub(regex, replace, tmp)
    regex = ',[ ]?"nodeid":[ ]?null'
    return re.sub(regex, replace, tmp)


def cross_check(raw_ext_query, be_query):
    ext_query = orjson.loads(replace_nodeid(PrismQuery(**raw_ext_query).json()))
    be_query = orjson.loads(replace_nodeid(be_query))
    assert ext_query == be_query


import pytest

from prism._common.const import FillnaMethodType
from prism._utils.validate_utils import _validate_args
from prism._utils.exceptions import PrismTypeError, PrismValueError


@_validate_args
def fillna(
        method: str,
        n: int = 0,
        value=None) -> FillnaMethodType:
    """
    valid arguments:
        'backfill', 'bfill', 'pad', 'ffill'
    """
    return method


def test_fillna_param():
    assert fillna(None, value=1, n=1) is None

    # check type
    # assert isinstance(fillna('backfill', n=1), FillnaMethodType)

    # # valid fillna param
    # assert fillna('backfill', n=1) == FillnaMethodType.BACKFILL
    # assert fillna('bfill', n=2) == FillnaMethodType.BFILL
    # assert fillna('pad', n=3) == FillnaMethodType.PAD
    # assert fillna('ffill', n=4) == FillnaMethodType.FFILL


def test_wrong_fillna_param():
    pass
    # wrong value
    # with pytest.raises(PrismValueError):
    #     assert fillna("Bfill", n=4) == FillnaMethodType.BFILL

    # with pytest.raises(PrismValueError):
    #     assert fillna("fill_zero", n=4)

    # # fillna() should only have one argument for either specifying a value or a method to fill the missing values
    # with pytest.raises(PrismValueError):
    #     assert fillna(value="1", method='bfill') == FillnaMethodType.BFILLs

    # # wrong type
    # with pytest.raises(PrismTypeError):
    #     assert fillna(method=123, n=4)


@_validate_args
def fill(
        method: str,
        n: int = 0,
        value=None) -> FillnaMethodType:
    return method


def test_fill_param():
    # check type
    # assert isinstance(fill('backfill', n=1), FillnaMethodType)

    # valid fill param
    assert fill('backfill', n=5) == FillnaMethodType.BACKFILL
    assert fill('bfill', n=6) == FillnaMethodType.BFILL
    assert fill('pad', n=7) == FillnaMethodType.PAD
    assert fill('ffill', n=8) == FillnaMethodType.FFILL


def test_wrong_fill_param():
    # wrong value
    with pytest.raises(Exception):
        assert fill("back fill", n=4) == FillnaMethodType.BACKFILL

    # with pytest.raises(Exception):
    #     assert fill(" ", n=4)

    # fill() can have argument value and method
    assert fill(value="1", method='bfill') == FillnaMethodType.BFILL

    # wrong type
    with pytest.raises(Exception):
        assert fill(method=1, n=4)



import pytest

from prism._utils.validate_utils import _validate_args
from prism._utils.exceptions import PrismTypeError


@_validate_args
def log(base: int):
    return True


def test_base_param_in_function_log():
    assert log(1)

    # valid base param
    assert log(2)
    assert log(base=4)


def test_wrong_base_param_in_function_log():
    # wrong value
    with pytest.raises(Exception):
        assert log(0)

    with pytest.raises(Exception):
        assert log(-1)

    # wrong type
    with pytest.raises(PrismTypeError):
        assert log("1")

    with pytest.raises(PrismTypeError):
        assert log("info")


@_validate_args
def sample_function(base: int):
    return True


def test_base_param_in_sample_function():
    # all integers are valid
    assert sample_function(0)
    assert sample_function(-2)
    assert sample_function(2)
    assert sample_function(base=4)


def test_wrong_base_param_in_sample_function():
    # wrong type
    with pytest.raises(PrismTypeError):
        assert sample_function("1")

    with pytest.raises(PrismTypeError):
        assert sample_function("info")



from ._req_builder import *
from ._req_builder import preference
from ._data import (
    financial,
    estimate,
    industryfin,
    market,
    precalculated,
    securitymaster,
    index,
    event,
    esg,
    transaction,
    company,
)
from ._task import (
    screen,
    export_data,
    factor_backtest,
    strategy_backtest,
)
from ._model import tcmodel, riskmodel
from ._fn import max, min, std, mean, var, sum



import pytest

from prism._utils.validate_utils import _validate_args
from prism._utils.exceptions import PrismTypeError, PrismValueError


@_validate_args
def get_settings(settings: dict):
    return True


def test_setting_param(mocker):
    mocker.patch("prism._common.const.PreferenceType", ["mode", "size", "author"])

    # valid setting keys
    assert get_settings({"mode": "white"})
    assert get_settings({"size": "10px"})
    assert get_settings({"author": "prism"})


def test_wrong_setting_param(mocker):
    mocker.patch("prism._common.const.PreferenceType", ["mode"])

    # wrong value
    with pytest.raises(PrismValueError):
        assert get_settings({"unknown": ""})

    with pytest.raises(PrismValueError):
        assert get_settings({"my_mode": ""})

    with pytest.raises(PrismValueError):
        assert get_settings({"MODE": ""})

    # wrong type
    with pytest.raises(PrismTypeError):
        assert get_settings(["mode"])



import os
from glob import glob
from dotenv import dotenv_values
from setuptools import setup, find_packages


env = os.environ['ENV_STATE']
name = "prismstudio"
version = dotenv_values("prism/_common/.env").get("VERSION", "")[1:]

datafiles = [("prism/_common", ["prism/_common/.env"])]
if env == 'dev' or env == 'stg' or env == 'demo':
    name = name + '-' + env

setup(
    name=name,
    version=version,
    description="Python Extension for PrismStudio",
    author="Prism39",
    author_email="jmp@prism39.com",
    url="https://www.prism39.com",
    packages=find_packages(),
    license="Prism39 End User License",
    python_requires=">=3.8",
    data_files=datafiles,
    classifiers=[
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    install_requires=[
        "pandas==2.0.3",
        "pyarrow==12.0.0",
        "tqdm==4.64.0",
        "orjson==3.9.3",
        "ipywidgets==7.7.1",
        "requests==2.26.0",
        "setuptools==58.1.0",
        "wheel==0.37.1",
        "twine==4.0.1",
        "pydantic[dotenv]==1.10.2",
        "pyzmq==24.0.1",
        "python-dotenv==1.0.0"
    ],
    include_package_data=True
)